#!/usr/bin/env python
# manifold.py
# Produce images for manifold documentation.
#
# Author:  Benjamin Bengfort
# Created: Sat May 12 11:26:18 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: manifold.py [] benjamin@bengfort.com $

"""
Produce images for manifold documentation.
"""

##########################################################################
# Imports
##########################################################################

import os
import matplotlib.pyplot as plt

from sklearn import datasets
from sklearn.pipeline import Pipeline
from sklearn.feature_selection import f_classif
from sklearn.feature_selection import SelectKBest

from yellowbrick.datasets import load_occupancy, load_concrete
from yellowbrick.features.manifold import (
    Manifold,
    MANIFOLD_ALGORITHMS,
    manifold_embedding,
)

SKIP = (
    "ltsa",  # produces no result
    "hessian",  # errors because of matrix
    "mds",  # uses way too much memory
)


FIXTURES = os.path.normpath(
    os.path.join(os.path.dirname(__file__), "..", "..", "..", "examples", "data")
)


def dataset_example(
    dataset="occupancy", manifold="all", path="images/", quick=False, **kwargs
):
    if manifold == "all":
        if path is not None and not os.path.isdir(path):
            "please specify a directory to save examples to"

        for algorithm in MANIFOLD_ALGORITHMS:
            if algorithm in SKIP:
                continue

            print("generating {} {} manifold".format(dataset, algorithm))
            fpath = os.path.join(path, "{}_{}_manifold.png".format(dataset, algorithm))
            try:
                dataset_example(dataset, algorithm, fpath)
            except Exception as e:
                print(
                    "could not visualize {} manifold on {} data: {}".format(
                        algorithm, dataset, e
                    )
                )
                continue

        # Break here!
        return

    # Create single example

    _, ax = plt.subplots(figsize=(9, 6))

    # Checks for the dataset which user wants to use
    if dataset == "occupancy":
        X, y = load_occupancy()
    elif dataset == "concrete":
        X, y = load_concrete()
    else:
        raise Exception("unknown dataset '{}'".format(dataset))

    # Check if the quick method is called
    if quick:
        oz = manifold_embedding(X, y, manifold=manifold, show=False, **kwargs)
        oz.show(outpath=path)
    else:
        oz = Manifold(ax=ax, manifold=manifold, **kwargs)
        oz.fit_transform(X, y)
        oz.show(outpath=path)


def select_features_example(
    algorithm="isomap",
    path="images/occupancy_select_k_best_isomap_manifold.png",
    **kwargs
):
    _, ax = plt.subplots(figsize=(9, 6))

    X, y = load_occupancy()
    classes = ["unoccupied", "occupied"]

    model = Pipeline(
        [
            ("selectk", SelectKBest(k=3, score_func=f_classif)),
            ("viz", Manifold(ax=ax, manifold=algorithm, classes=classes, **kwargs)),
        ]
    )

    model.fit_transform(X, y)
    model.named_steps["viz"].show(outpath=path)


class SCurveExample(object):
    """
    Creates an S-curve example and multiple visualizations
    """

    def __init__(self, n_points=1000, random_state=42):
        self.X, self.y = datasets.samples_generator.make_s_curve(
            n_points, random_state=random_state
        )

    def _make_path(self, path, name):
        """
        Makes directories as needed
        """
        if not os.path.exists(path):
            os.mkdirs(path)

        if os.path.isdir(path):
            return os.path.join(path, name)

        return path

    def plot_original_3d(self, path="images"):
        """
        Plot the original data in 3-dimensional space
        """
        raise NotImplementedError("nyi")

    def plot_manifold_embedding(self, algorithm="lle", path="images"):
        """
        Draw the manifold embedding for the specified algorithm
        """
        _, ax = plt.subplots(figsize=(9, 6))
        path = self._make_path(path, "s_curve_{}_manifold.png".format(algorithm))

        oz = Manifold(ax=ax, manifold=algorithm, colors="nipy_spectral")

        oz.fit_transform(self.X, self.y)
        oz.show(outpath=path)

    def plot_all_manifolds(self, path="images"):
        """
        Plot all s-curve examples
        """
        for algorithm in MANIFOLD_ALGORITHMS:
            self.plot_manifold_embedding(algorithm)


##########################################################################
# Main Method
##########################################################################

if __name__ == "__main__":
    # curve = SCurveExample()
    # curve.plot_all_manifolds()
    dataset_example("concrete", "tsne", path="images/concrete_tsne_manifold.png")
    dataset_example(
        "occupancy",
        "tsne",
        classes=["unoccupied", "occupied"],
        path="images/occupancy_tsne_manifold.png",
    )
    dataset_example(
        "concrete", "isomap", path="images/concrete_isomap_manifold.png", n_neighbors=10
    )
    dataset_example(
        "concrete",
        "isomap",
        target="continuous",
        path="images/manifold_quick_method.png",
        n_neighbors=10,
        quick=True,
    )
    select_features_example(algorithm="isomap", n_neighbors=10)



# tests.test_features.test_projection
# Test the base ProjectionVisualizer drawing functionality
#
# Author:   Naresh Bachwani
# Created:  Wed Jul 17 09:53:07 2019 -0400
#
# Copyright (C) 2019 the scikit-yb developers.
# For license information, see LICENSE.txt
#
# ID: test_projection.py [21eb9d2] 43993586+naresh-bachwani@users.noreply.github.com $

"""
Test the base ProjectionVisualizer drawing functionality
"""

##########################################################################
## Imports
##########################################################################

import pytest
import numpy.testing as npt
import matplotlib.pyplot as plt

from yellowbrick.features.projection import *
from yellowbrick.exceptions import YellowbrickValueError

from tests.base import VisualTestCase
from unittest import mock

from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

##########################################################################
## MockVisualizer
##########################################################################


class MockVisualizer(ProjectionVisualizer):
    """
    The MockVisualizer implements the ProjectionVisualizer interface using
    PCA as an internal transformer. This visualizer is used to directly test
    how subclasses interact with the ProjectionVisualizer base class.
    """

    def __init__(
        self,
        ax=None,
        features=None,
        classes=None,
        colors=None,
        colormap=None,
        target_type="auto",
        projection=2,
        alpha=0.75,
        colorbar=True,
        **kwargs
    ):

        super(MockVisualizer, self).__init__(
            ax=ax,
            features=features,
            classes=classes,
            colors=colors,
            colormap=colormap,
            target_type=target_type,
            projection=projection,
            alpha=alpha,
            colorbar=colorbar,
            **kwargs
        )

        self.pca_transformer = Pipeline(
            [
                ("scale", StandardScaler()),
                ("pca", PCA(self.projection, random_state=2019)),
            ]
        )

    def fit(self, X, y=None):
        super(MockVisualizer, self).fit(X, y)
        self.pca_transformer.fit(X)
        return self

    def transform(self, X, y=None):
        try:
            Xp = self.pca_transformer.transform(X)
        except AttributeError as e:
            raise AttributeError(str(e) + " try using fit_transform instead.")
        self.draw(Xp, y)
        return Xp


##########################################################################
## ProjectionVisualizer Tests
##########################################################################


@pytest.mark.usefixtures("discrete", "continuous")
class TestProjectionVisualizer(VisualTestCase):
    """
    Test the ProjectionVisualizer base class
    """

    def test_discrete_plot(self):
        """
        Test the visualizer with discrete target.
        """
        X, y = self.discrete
        classes = ["a", "b", "c", "d", "e"]
        visualizer = MockVisualizer(projection=2, colormap="plasma", classes=classes)
        X_prime = visualizer.fit_transform(X, y)
        npt.assert_array_equal(visualizer.classes_, classes)
        visualizer.finalize()
        self.assert_images_similar(visualizer)
        assert X_prime.shape == (self.discrete.X.shape[0], 2)

    def test_continuous_plot(self):
        """
        Test the visualizer with continuous target.
        """
        X, y = self.continuous
        visualizer = MockVisualizer(projection="2d")
        visualizer.fit_transform(X, y)
        visualizer.finalize()
        visualizer.cax.set_yticklabels([])
        self.assert_images_similar(visualizer)

    def test_continuous_when_target_discrete(self):
        """
        Ensure user can override discrete target_type by specifying continuous
        """
        _, ax = plt.subplots()
        X, y = self.discrete
        visualizer = MockVisualizer(
            ax=ax, projection="2D", target_type="continuous", colormap="cool"
        )
        visualizer.fit(X, y)
        visualizer.transform(X, y)
        visualizer.finalize()
        visualizer.cax.set_yticklabels([])
        self.assert_images_similar(visualizer)

    def test_single_plot(self):
        """
        Assert single color plot when y is not specified
        """
        X, y = self.discrete
        visualizer = MockVisualizer(projection=2, colormap="plasma")
        visualizer.fit_transform(X)
        visualizer.finalize()
        self.assert_images_similar(visualizer)

    def test_discrete_3d(self):
        """
        Test visualizer for 3 dimensional discrete plots
        """
        X, y = self.discrete

        classes = ["a", "b", "c", "d", "e"]
        colors = ["r", "b", "g", "m", "c"]
        visualizer = MockVisualizer(projection=3, colors=colors, classes=classes)
        visualizer.fit_transform(X, y)
        npt.assert_array_equal(visualizer.classes_, classes)
        visualizer.finalize()
        self.assert_images_similar(visualizer)

    def test_3d_continuous_plot(self):
        """
        Tests visualizer for 3 dimensional continuous plots
        """
        X, y = self.continuous
        visualizer = MockVisualizer(projection="3D")
        visualizer.fit_transform(X, y)
        visualizer.finalize()
        visualizer.cbar.set_ticks([])
        self.assert_images_similar(visualizer)

    def test_alpha_param(self):
        """
        Ensure that the alpha parameter modifies opacity
        """
        # Instantiate a prediction error plot, provide custom alpha
        X, y = self.discrete
        params = {"alpha": 0.3, "projection": 2}
        visualizer = MockVisualizer(**params)
        visualizer.ax = mock.MagicMock()
        visualizer.fit(X, y)
        visualizer.transform(X, y)

        assert visualizer.alpha == 0.3

        # Test that alpha was passed to internal matplotlib scatterplot
        _, scatter_kwargs = visualizer.ax.scatter.call_args
        assert "alpha" in scatter_kwargs
        assert scatter_kwargs["alpha"] == 0.3

    # Check Errors
    @pytest.mark.parametrize("projection", ["4D", 1, "100d", 0])
    def test_wrong_projection_dimensions(self, projection):
        """
        Validate projection hyperparameter
        """
        msg = "Projection dimensions must be either 2 or 3"
        with pytest.raises(YellowbrickValueError, match=msg):
            MockVisualizer(projection=projection)

    def test_target_not_label_encoded(self):
        """
        Assert label encoding mismatch with y raises exception
        """
        X, y = self.discrete
        # Multiply every element by 10 to make non-label encoded
        y = y * 10
        visualizer = MockVisualizer()
        msg = "Target needs to be label encoded."
        with pytest.raises(YellowbrickValueError, match=msg):
            visualizer.fit_transform(X, y)

    @pytest.mark.parametrize("dataset", ("discrete", "continuous"))
    def test_y_required_for_discrete_and_continuous(self, dataset):
        """
        Assert error is raised when y is not passed to transform
        """
        X, y = getattr(self, dataset)
        visualizer = MockVisualizer()
        visualizer.fit(X, y)

        msg = "y is required for {} target".format(dataset)
        with pytest.raises(YellowbrickValueError, match=msg):
            visualizer.transform(X)

    def test_colorbar_false(self):
        """
        Test that colorbar equals false works correctly
        """
        visualizer = MockVisualizer(colorbar=False, colormap="YlOrRd")
        visualizer.fit_transform(*self.continuous)
        visualizer.finalize()

        self.assert_images_similar(visualizer)



#!/usr/bin/env python3

import os
import sys
import argparse
import numpy as np
import matplotlib.pyplot as plt

from yellowbrick.regressor import AlphaSelection
from yellowbrick.regressor import ResidualsPlot
from yellowbrick.regressor import PredictionError
from yellowbrick.cluster import KElbowVisualizer
from yellowbrick.cluster import SilhouetteVisualizer
from yellowbrick.cluster import InterclusterDistance
from yellowbrick.classifier import PrecisionRecallCurve
from yellowbrick.classifier import ClassPredictionError
from yellowbrick.classifier import DiscriminationThreshold
from yellowbrick.datasets import load_spam, load_concrete, load_game


from sklearn.cluster import KMeans
from sklearn.datasets import make_blobs
from sklearn.naive_bayes import GaussianNB
from sklearn.preprocessing import OneHotEncoder
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression, Lasso, LassoCV, Ridge


FIGSIZE = (20, 4)

IMAGES = os.path.normpath(os.path.join(os.path.dirname(__file__), ".."))
YB_LOGO_PATH = os.path.join(IMAGES, "yb-fc.png")


def tts_plot(viz, X, y, test_size=0.20, random_state=42, score=True, finalize=True):
    """
    Helper function to plot model visualizers with train_test_split
    """
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, random_state=random_state
    )

    viz.fit(X_train, y_train)
    if score:
        viz.score(X_test, y_test)
    if finalize:
        viz.finalize()

    return viz


def class_prediction_error(ax=None):
    data = load_game(return_dataset=True)
    X, y = data.to_numpy()

    X = OneHotEncoder().fit_transform(X).toarray()

    viz = ClassPredictionError(GaussianNB(), ax=ax)
    return tts_plot(viz, X, y)


def confusion_matrix(ax=None):
    data = load_spam(return_dataset=True)
    X, y = data.to_pandas()

    viz = PrecisionRecallCurve(LogisticRegression(), ax=ax)
    return tts_plot(viz, X, y)


def discrimination_threshold(ax=None):
    data = load_spam(return_dataset=True)
    X, y = data.to_pandas()

    viz = DiscriminationThreshold(RandomForestClassifier(n_estimators=10), ax=ax)
    return tts_plot(viz, X, y, score=False)


def classification_visualizers(saveto=None):
    _, (axa, axb, axc) = plt.subplots(nrows=1, ncols=3, figsize=FIGSIZE)

    class_prediction_error(axa)
    confusion_matrix(axb)
    discrimination_threshold(axc)

    plt.tight_layout(pad=1.5)

    if saveto is not None:
        plt.savefig(saveto)
    else:
        plt.show()


def residuals_plot(ax=None):
    data = load_concrete(return_dataset=True)
    X, y = data.to_pandas()

    viz = ResidualsPlot(Ridge(), ax=ax)
    return tts_plot(viz, X, y)


def prediction_error(ax=None):
    data = load_concrete(return_dataset=True)
    X, y = data.to_pandas()

    viz = PredictionError(Lasso(), ax=ax)
    return tts_plot(viz, X, y)


def alpha_selection(ax=None):
    data = load_concrete(return_dataset=True)
    X, y = data.to_pandas()

    alphas = np.logspace(-10, 1, 400)
    viz = AlphaSelection(LassoCV(alphas=alphas), ax=ax)
    return tts_plot(viz, X, y)


def regression_visualizers(saveto=None):
    _, (axa, axb, axc) = plt.subplots(nrows=1, ncols=3, figsize=FIGSIZE)

    residuals_plot(axa)
    prediction_error(axb)
    alpha_selection(axc)

    plt.tight_layout(pad=1.5)

    if saveto is not None:
        plt.savefig(saveto)
    else:
        plt.show()


def intercluster_distance(ax=None):
    X, y = make_blobs(centers=12, n_samples=1000, n_features=16, shuffle=True)

    viz = InterclusterDistance(KMeans(9), ax=ax)
    viz.fit(X)
    viz.finalize()

    return viz


def k_elbow(ax=None):
    X, y = make_blobs(centers=12, n_samples=1000, n_features=16, shuffle=True)

    viz = KElbowVisualizer(KMeans(), k=(4, 12), ax=ax, locate_elbow=False)
    viz.fit(X)
    viz.finalize()

    return viz


def silhouette(ax=None):
    X, y = make_blobs(centers=12, n_samples=1000, n_features=16, shuffle=True)

    viz = SilhouetteVisualizer(KMeans(9), ax=ax)
    viz.fit(X)
    viz.finalize()

    return viz


def clustering_visualizers(saveto=None):
    _, (axa, axb, axc) = plt.subplots(nrows=1, ncols=3, figsize=FIGSIZE)

    intercluster_distance(axa)
    k_elbow(axb)
    silhouette(axc).ax.get_legend().remove()

    plt.tight_layout(pad=1.5)

    if saveto is not None:
        plt.savefig(saveto)
    else:
        plt.show()


def yb_logo(path=YB_LOGO_PATH, ax=None):
    """
    Reads the YB image logo from the specified path and writes it to the axes.
    """
    # Load image
    with open(path, "rb") as fobj:
        img = plt.imread(fobj, format="png")

    if ax is None:
        _, ax = plt.subplots()

    # Draw image
    ax.imshow(img, interpolation="nearest")

    # Remove spines, ticks, grid, and other marks
    ax.grid(False)
    ax.set_xticks([])
    ax.set_yticks([])
    for pos in ["right", "top", "bottom", "left"]:
        ax.spines[pos].set_visible(False)

    return ax


def full_image(saveto=None, center_logo=False):
    _, axes = plt.subplots(nrows=3, ncols=3, figsize=(21, 14))

    # Top row: classifiers
    class_prediction_error(axes[0][0])
    confusion_matrix(axes[0][1])
    discrimination_threshold(axes[0][2])

    # Middle row: regressors
    residuals_plot(axes[1][0])
    alpha_selection(axes[1][2])

    if center_logo:
        yb_logo(ax=axes[1][1])
    else:
        prediction_error(axes[1][1])

    # Bottom row: clusterers
    intercluster_distance(axes[2][0])
    k_elbow(axes[2][1])
    silhouette(axes[2][2]).ax.get_legend().remove()

    plt.tight_layout(pad=1.5)

    if saveto is not None:
        plt.savefig(saveto)
    else:
        plt.show()


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="generates images for the README.md banner"
    )

    parser.add_argument(
        "-c",
        "--classifiers",
        type=str,
        metavar="PATH",
        default="classifiers.png",
        help="path to save the classifiers banner image",
    )

    parser.add_argument(
        "-r",
        "--regressors",
        type=str,
        metavar="PATH",
        default="regressors.png",
        help="path to save the regressors banner image",
    )

    parser.add_argument(
        "-C",
        "--clusterers",
        type=str,
        metavar="PATH",
        default="clusterers.png",
        help="path to save the clusterers banner image",
    )

    parser.add_argument(
        "-b",
        "--banner",
        type=str,
        metavar="PATH",
        default="",
        help="make full banner image and save to disk",
    )

    parser.add_argument(
        "-y",
        "--yb",
        action="store_true",
        help="replace middle image of banner with logo",
    )

    args = parser.parse_args()

    if args.banner:
        full_image(args.banner, args.yb)
        sys.exit(0)

    if args.classifiers:
        classification_visualizers(args.classifiers)

    if args.regressors:
        regression_visualizers(args.regressors)

    if args.clusterers:
        clustering_visualizers(args.clusterers)



# tests.test_model_selection.test_validation_curve
# Tests for the ValidationCurve visualizer
#
# Author:  Benjamin Bengfort
# Created: Sat Mar 31 06:25:05 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_validation_curve.py [c5355ee] benjamin@bengfort.com $

"""
Tests for the ValidationCurve visualizer
"""

##########################################################################
# Imports
##########################################################################

import sys
import pytest
import numpy as np

from unittest.mock import patch
from tests.base import VisualTestCase

from sklearn.svm import SVC
from sklearn.naive_bayes import BernoulliNB
from sklearn.tree import DecisionTreeRegressor
from sklearn.preprocessing import OneHotEncoder
from sklearn.neighbors import KNeighborsClassifier
from sklearn.model_selection import ShuffleSplit, StratifiedKFold

from yellowbrick.datasets import load_mushroom
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.model_selection.validation_curve import *


try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
# Test Cases
##########################################################################


@pytest.mark.usefixtures("classification", "regression", "clusters")
class TestValidationCurve(VisualTestCase):
    """
    Test the ValidationCurve visualizer
    """

    @patch.object(ValidationCurve, "draw")
    def test_fit(self, mock_draw):
        """
        Assert that fit returns self and creates expected properties
        """
        X, y = self.classification
        params = (
            "train_scores_",
            "train_scores_mean_",
            "train_scores_std_",
            "test_scores_",
            "test_scores_mean_",
            "test_scores_std_",
        )

        oz = ValidationCurve(
            SVC(), param_name="gamma", param_range=np.logspace(-6, -1, 5)
        )

        for param in params:
            assert not hasattr(oz, param)

        assert oz.fit(X, y) is oz
        mock_draw.assert_called_once()

        for param in params:
            assert hasattr(oz, param)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_classifier(self):
        """
        Test image closeness on a classification dataset with kNN
        """
        X, y = self.classification

        cv = ShuffleSplit(3, random_state=288)
        param_range = np.arange(3, 10)

        oz = ValidationCurve(
            KNeighborsClassifier(),
            param_name="n_neighbors",
            param_range=param_range,
            cv=cv,
            scoring="f1_weighted",
        )

        oz.fit(X, y)
        oz.finalize()

        self.assert_images_similar(oz)

    def test_regression(self):
        """
        Test image closeness on a regression dataset with a DecisionTree
        """
        X, y = self.regression

        cv = ShuffleSplit(3, random_state=938)
        param_range = np.arange(3, 10)

        oz = ValidationCurve(
            DecisionTreeRegressor(random_state=23),
            param_name="max_depth",
            param_range=param_range,
            cv=cv,
            scoring="r2",
        )

        oz.fit(X, y)
        oz.finalize()

        self.assert_images_similar(oz, tol=12.0)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_quick_method(self):
        """
        Test validation curve quick method with image closeness on SVC
        """
        X, y = self.classification

        pr = np.logspace(-6, -1, 3)
        cv = ShuffleSplit(n_splits=5, test_size=0.2, random_state=321)
        viz = validation_curve(
            SVC(), X, y, logx=True, param_name="gamma",
            param_range=pr, cv=cv, show=False
        )

        self.assert_images_similar(viz)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_pandas_integration(self):
        """
        Test on mushroom dataset with pandas DataFrame and Series and NB
        """
        data = load_mushroom(return_dataset=True)
        X, y = data.to_pandas()

        X = pd.get_dummies(X)

        assert isinstance(X, pd.DataFrame)
        assert isinstance(y, pd.Series)

        cv = StratifiedKFold(n_splits=2, shuffle=True, random_state=11)
        pr = np.linspace(0.1, 3.0, 6)
        oz = ValidationCurve(BernoulliNB(), cv=cv, param_range=pr, param_name="alpha")
        oz.fit(X, y)
        oz.finalize()

        self.assert_images_similar(oz)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_numpy_integration(self):
        """
        Test on mushroom dataset with NumPy arrays
        """
        data = load_mushroom(return_dataset=True)
        X, y = data.to_numpy()

        X = OneHotEncoder().fit_transform(X).toarray()

        cv = StratifiedKFold(n_splits=2, shuffle=True, random_state=11)
        pr = np.linspace(0.1, 3.0, 6)
        oz = ValidationCurve(BernoulliNB(), cv=cv, param_range=pr, param_name="alpha")
        oz.fit(X, y)
        oz.finalize()

        self.assert_images_similar(oz)

    @patch.object(ValidationCurve, "draw")
    def test_reshape_scores(self, mock_draw):
        """
        Test supplying an alternate CV methodology and train_sizes
        """
        X, y = self.classification

        pr = np.logspace(-6, -1, 3)
        cv = ShuffleSplit(n_splits=5, test_size=0.2, random_state=14)
        oz = ValidationCurve(SVC(), param_name="gamma", param_range=pr, cv=cv)
        oz.fit(X, y)

        assert oz.train_scores_.shape == (3, 5)
        assert oz.test_scores_.shape == (3, 5)

    def test_bad_train_sizes(self):
        """
        Test learning curve with bad input for training size.
        """
        with pytest.raises(YellowbrickValueError):
            ValidationCurve(SVC(), param_name="gamma", param_range=100)



# tests.test_utils.test_kneed
# A port of the tests for knee-point detection package, kneed.
#
# Author:   Kevin Arvai
# Author:   Pradeep Singh
# Created:  Mon Apr 23 01:29:18 2019 -0400
#
# Copyright (C) 2017 Kevin Arvai
# All rights reserved.
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list
# of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or other
# materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may
# be used to endorse or promote products derived from this software without specific
# prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# ID: test_kneed.py [] pswaldia@no-reply.github.com $

"""
This package contains a port of the tests for knee-point detection package, kneed, by
Kevin Arvai and hosted at https://github.com/arvkevi/kneed. This port is maintained
with permission by the Yellowbrick contributors.
"""

import pytest
import matplotlib.pyplot as plt
import numpy as np
from yellowbrick.utils.kneed import KneeLocator

x = np.arange(0, 10)
y_convex_inc = np.array([1, 2, 3, 4, 5, 10, 15, 20, 40, 100])
y_convex_dec = np.array(y_convex_inc[::-1])
y_concave_dec = np.array(100 - y_convex_inc)
y_concave_inc = np.array(100 - y_convex_dec)


def test_concave_increasing():
    """Tests that a correct knee point is detected in
    curve having concave and increasing nature."""
    kn = KneeLocator(
        x, y_concave_inc, curve_nature="concave", curve_direction="increasing"
    )
    assert kn.knee == 2


def test_concave_decreasing():
    """Tests that a correct knee point is detected in
    curve having concave and decreasing nature."""
    kn = KneeLocator(
        x, y_concave_dec, curve_nature="concave", curve_direction="decreasing"
    )
    assert kn.knee == 7


def test_convex_increasing():
    """Tests that a correct knee point is detected in
    curve having convex and increasing nature."""
    kn = KneeLocator(
        x, y_convex_inc, curve_nature="convex", curve_direction="increasing"
    )
    assert kn.knee == 7


def test_convex_decreasing():
    """Tests that a correct knee point is detected in
    curve having convex and decreasing nature."""
    kn = KneeLocator(
        x, y_convex_dec, curve_nature="convex", curve_direction="decreasing"
    )
    assert kn.knee == 2


def test_concave_increasing_truncated():
    """Tests that a correct knee point is detected in
    curve having truncated concave increasing nature"""
    kn = KneeLocator(
        x[:-3] / 10,
        y_concave_inc[:-3] / 10,
        curve_nature="concave",
        curve_direction="increasing",
    )
    assert kn.knee == 0.2


def test_concave_decreasing_truncated():
    """Tests that a correct knee point is detected in
    curve having truncated concave decreasing nature"""
    kn = KneeLocator(
        x[:-3] / 10,
        y_concave_dec[:-3] / 10,
        curve_nature="concave",
        curve_direction="decreasing",
    )
    assert kn.knee == 0.4


def test_convex_increasing_truncated():
    """Tests that a correct knee point is detected in
    curve having truncated convex increasing nature"""
    kn = KneeLocator(
        x[:-3] / 10,
        y_convex_inc[:-3] / 10,
        curve_nature="convex",
        curve_direction="increasing",
    )
    assert kn.knee == 0.4


def test_convex_decreasing_truncated():
    """Tests that a correct knee point is detected in
    curve having truncated convex decreasing nature"""
    kn = KneeLocator(
        x[:-3] / 10,
        y_convex_dec[:-3] / 10,
        curve_nature="convex",
        curve_direction="decreasing",
    )
    assert kn.knee == 0.2


def test_x_equals_y():
    """Test that a runtime warning is raised when no maxima are found"""
    x = range(10)
    y = [1] * len(x)
    with pytest.warns(RuntimeWarning):
        KneeLocator(x, y)


@pytest.mark.parametrize("online, expected", [(True, 482), (False, 22)])
def test_gamma_online_offline(online, expected):
    """Tests online and offline knee detection.
    Notable that a large number of samples are highly sensitive to S parameter
    """
    np.random.seed(23)
    n = 1000
    x = range(1, n + 1)
    y = sorted(np.random.gamma(0.5, 1.0, n), reverse=True)
    kl = KneeLocator(x, y, curve_nature="convex", curve_direction="decreasing", online=online)
    assert kl.knee == expected


def test_properties():
    """Tests that elbow and knee can be used interchangeably."""
    kn = KneeLocator(
        x, y_concave_inc, curve_nature="concave", curve_direction="increasing"
    )
    assert kn.knee == kn.elbow
    assert kn.norm_knee == kn.norm_elbow
    # pytest compares all elements in each list.
    assert kn.all_knees == kn.all_elbows
    assert kn.all_norm_knees == kn.all_norm_elbows


def test_plot_knee_normalized():
    """Test that plotting is functional"""
    with np.errstate(divide="ignore"):
        x = np.linspace(0.0, 1, 10)
        y = np.true_divide(-1, x + 0.1) + 5
    kl = KneeLocator(x, y, S=1.0, curve_nature="concave")
    num_figures_before = plt.gcf().number
    kl.plot_knee_normalized()
    num_figures_after = plt.gcf().number
    assert num_figures_before < num_figures_after


def test_plot_knee():
    """Test that plotting is functional"""
    with np.errstate(divide="ignore"):
        x = np.linspace(0.0, 1, 10)
        y = np.true_divide(-1, x + 0.1) + 5
    kl = KneeLocator(x, y, S=1.0, curve_nature="concave")
    num_figures_before = plt.gcf().number
    kl.plot_knee()
    num_figures_after = plt.gcf().number
    assert num_figures_before < num_figures_after


def test_y():
    """Test the y value"""
    with np.errstate(divide="ignore"):
        x = np.linspace(0.0, 1, 10)
        y = np.true_divide(-1, x + 0.1) + 5
    kl = KneeLocator(x, y, S=1.0, curve_nature="concave")
    assert kl.knee_y == pytest.approx(1.897, 0.03)
    assert kl.all_knees_y[0] == pytest.approx(1.897, 0.03)
    assert kl.norm_knee_y == pytest.approx(0.758, 0.03)
    assert kl.all_norm_knees_y[0] == pytest.approx(0.758, 0.03)

    assert kl.elbow_y == pytest.approx(1.897, 0.03)
    assert kl.all_elbows_y[0] == pytest.approx(1.897, 0.03)
    assert kl.norm_elbow_y == pytest.approx(0.758, 0.03)
    assert kl.all_norm_elbows_y[0] == pytest.approx(0.758, 0.03)



# tests.test_features.test_jointplot
# Test the JointPlot Visualizer
#
# Author:   Prema Damodaran Roman
# Created:  Mon Apr 10 21:00:54 2017 -0400
#
# Copyright (C) 2017 The scikit-yb developers.
# For license information, see LICENSE.txt
#
# ID: test_jointplot.py [9e008b0] pdamodaran@users.noreply.github.com $

"""
Test joint plot visualization methods.

These tests work differently depending on what version of matplotlib is
installed. If version 2.0.2 or greater is installed, then most tests will
execute, otherwise the histogram tests will skip and only the warning will
be tested.
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest
import numpy as np

from functools import partial
from unittest.mock import patch, MagicMock

from sklearn.datasets import make_classification, make_regression

from tests.base import IS_WINDOWS_OR_CONDA, VisualTestCase
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.features.jointplot import *
from ..fixtures import Dataset

try:
    # Only available in Matplotlib >= 2.0.2
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    make_axes_locatable = None

try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
## Fixtures
##########################################################################

# Random numpy array generators
rand1d = partial(np.random.rand, 120)
rand2col = partial(np.random.rand, 120, 2)
rand3col = partial(np.random.rand, 120, 3)


@pytest.fixture(scope="class")
def discrete(request):
    """
    Creates a simple 2-column dataset with a discrete target.
    """
    X, y = make_classification(
        n_samples=120,
        n_features=2,
        n_informative=2,
        n_redundant=0,
        n_classes=3,
        n_clusters_per_class=1,
        random_state=2221,
    )

    request.cls.discrete = Dataset(X, y)


@pytest.fixture(scope="class")
def continuous(request):
    """
    Creates a simple 2-column dataset with a continuous target.
    """
    X, y = make_regression(n_samples=120, n_features=2, random_state=1112)

    request.cls.continuous = Dataset(X, y)


##########################################################################
## JointPlot Tests
##########################################################################


@pytest.mark.usefixtures("discrete", "continuous")
class TestJointPlotNoHistogram(VisualTestCase):
    """
    Test the JointPlot visualizer without histograms
    """

    def test_invalid_columns_values(self):
        """
        Assert invalid columns arguments raise exception
        """
        with pytest.raises(YellowbrickValueError, match="invalid for joint plot"):
            JointPlot(columns=["a", "b", "c"], hist=False)

    def test_invalid_correlation_values(self):
        """
        Assert invalid correlation arguments raise an exception
        """
        with pytest.raises(YellowbrickValueError, match="invalid correlation method"):
            JointPlot(correlation="foo", hist=False)

    def test_invalid_kind_values(self):
        """
        Assert invalid kind arguments raise exception
        """
        for bad_kind in ("foo", None, 123):
            with pytest.raises(YellowbrickValueError, match="invalid joint plot kind"):
                JointPlot(kind=bad_kind, hist=False)

    def test_invalid_hist_values(self):
        """
        Assert invalid hist arguments raise exception
        """
        for bad_hist in ("foo", 123):
            with pytest.raises(
                YellowbrickValueError, match="invalid argument for hist"
            ):
                JointPlot(hist=bad_hist)

    def test_no_haxes(self):
        """
        Test that xhax and yhax are not available
        """
        oz = JointPlot(hist=False)
        with pytest.raises(AttributeError, match="histogram for the X axis"):
            oz.xhax

        with pytest.raises(AttributeError, match="histogram for the Y axis"):
            oz.yhax

    @patch("yellowbrick.features.jointplot.plt")
    def test_correlation(self, mplt):
        """
        Test correlation is correctly computed
        """
        x = self.discrete.X[:, 0]
        y = self.discrete.X[:, 1]

        cases = (
            ("pearson", -0.3847799883805261),
            ("spearman", -0.37301201472324463),
            ("covariance", -0.5535440619953924),
            ("kendalltau", -0.2504201680672269),
        )

        for alg, expected in cases:
            oz = JointPlot(hist=False, correlation=alg, columns=None)
            oz.ax = MagicMock()
            oz.fit(x, y)

            assert hasattr(oz, "corr_")
            assert oz.corr_ == pytest.approx(
                expected
            ), "{} not computed correctly".format(alg)

    def test_columns_none_invalid_x(self):
        """
        When self.columns=None validate X and y
        """
        bad_kws = (
            {"X": rand1d(), "y": None},
            {"X": rand3col(), "y": None},
            {"X": rand2col(), "y": rand1d()},
            {"X": rand3col(), "y": rand1d()},
            {"X": rand1d(), "y": rand2col()},
        )

        for kws in bad_kws:
            oz = JointPlot(columns=None, hist=False)
            with pytest.raises(
                YellowbrickValueError, match="when self.columns is None"
            ):
                oz.fit(**kws)

    def test_columns_none_x_y(self):
        """
        When self.columns=None image similarity with valid X and y
        """
        oz = JointPlot(hist=False, columns=None)
        assert oz.fit(self.discrete.X[:, 0], self.discrete.y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        # Appveyor and Linux conda fail due to non-text-based differences
        self.assert_images_similar(oz, tol=2.5)

    def test_columns_none_x(self):
        """
        When self.columns=None image similarity with valid X, no y
        """
        oz = JointPlot(hist=False, columns=None)
        assert oz.fit(self.discrete.X) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        tol = (
            4.0 if sys.platform == "win32" else 0.01
        )  # Fails on AppVeyor with RMS 3.941
        self.assert_images_similar(oz, tol=tol)

    def test_columns_single_index_no_y(self):
        """
        When self.columns=int or str y must not be None
        """
        oz = JointPlot(columns="foo", hist=False)
        with pytest.raises(YellowbrickValueError, match="y must be specified"):
            oz.fit(rand2col(), y=None)

    def test_columns_single_invalid_index_numpy(self):
        """
        When self.columns=int validate the index in X
        """
        oz = JointPlot(columns=2, hist=False)
        with pytest.raises(IndexError, match="could not index column '2' into type"):
            oz.fit(self.continuous.X, self.continuous.y)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_columns_single_invalid_index_pandas(self):
        """
        When self.columns=str validate the index in X
        """
        oz = JointPlot(columns="foo", hist=False)
        X = pd.DataFrame(self.continuous.X, columns=["a", "b"])
        y = pd.Series(self.continuous.y)

        with pytest.raises(IndexError, match="could not index column 'foo' into type"):
            oz.fit(X, y)

    def test_columns_single_int_index_numpy(self):
        """
        When self.columns=int image similarity on numpy dataset
        """
        oz = JointPlot(columns=1, hist=False)
        assert oz.fit(self.continuous.X, self.continuous.y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        # Appveyor and Linux conda failed based on non-text-based differences
        self.assert_images_similar(oz, tol=5)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_columns_single_str_index_pandas(self):
        """
        When self.columns=str image similarity on pandas dataset
        """
        oz = JointPlot(columns="a", hist=False)
        X = pd.DataFrame(self.continuous.X, columns=["a", "b"])
        y = pd.Series(self.continuous.y)
        assert oz.fit(X, y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        # Appveyor and Linux conda failed based on non-text-based differences
        self.assert_images_similar(oz, tol=5.5)

    def test_columns_double_int_index_numpy_no_y(self):
        """
        When self.columns=[int, int] image similarity on numpy dataset no y
        """
        oz = JointPlot(columns=[0, 1], hist=False)
        assert oz.fit(self.discrete.X, y=None) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        tol = (
            4.0 if sys.platform == "win32" else 0.01
        )  # Fails on AppVeyor with RMS 3.941
        self.assert_images_similar(oz, tol=tol)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_columns_double_str_index_pandas_no_y(self):
        """
        When self.columns=[str, str] image similarity on pandas dataset no y
        """
        oz = JointPlot(columns=["a", "b"], hist=False)
        X = pd.DataFrame(self.continuous.X, columns=["a", "b"])
        assert oz.fit(X, y=None) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        tol = (
            4.0 if sys.platform == "win32" else 0.01
        )  # Fails on AppVeyor with RMS 3.911
        self.assert_images_similar(oz, tol=tol)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_columns_double_index_discrete_y(self):
        """
        When self.columns=[str, str] on DataFrame with discrete y
        """
        oz = JointPlot(columns=["a", "b"], hist=False)
        X = pd.DataFrame(self.discrete.X, columns=["a", "b"])
        y = pd.Series(self.discrete.y)
        assert oz.fit(X, y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        tol = (
            4.0 if sys.platform == "win32" else 0.01
        )  # Fails on AppVeyor with RMS 3.940
        self.assert_images_similar(oz, tol=tol)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_columns_double_index_continuous_y(self):
        """
        When self.columns=[str, str] on DataFrame with continuous y
        """
        oz = JointPlot(columns=["a", "b"], hist=False)
        X = pd.DataFrame(self.continuous.X, columns=["a", "b"])
        y = pd.Series(self.continuous.y)
        assert oz.fit(X, y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        tol = (
            4.0 if sys.platform == "win32" else 0.01
        )  # Fails on AppVeyor with RMS 3.911
        self.assert_images_similar(oz, tol=tol)


@pytest.mark.skipif(
    make_axes_locatable is not None, reason="requires matplotlib <= 2.0.1"
)
def test_matplotlib_version_error():
    """
    Assert an exception is raised with incompatible matplotlib versions
    """
    with pytest.raises(YellowbrickValueError):
        JointPlot(hist=True)


@patch("yellowbrick.features.jointplot.make_axes_locatable", None)
def test_matplotlib_incompatibility():
    """
    Assert an exception is raised if make_axes_locatable is None
    """
    with pytest.raises(YellowbrickValueError):
        JointPlot(hist=True)


@pytest.mark.usefixtures("discrete", "continuous")
@pytest.mark.skipif(make_axes_locatable is None, reason="requires matplotlib >= 2.0.2")
class TestJointPlotHistogram(VisualTestCase):
    """
    Test the JointPlot visualizer with histograms
    """

    def test_haxes_available(self):
        """
        Test that xhax and yhax are available
        """
        oz = JointPlot(hist=True)
        assert oz.xhax is not None
        assert oz.yhax is not None

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_columns_none_x_y_hist(self):
        """
        When self.columns=None image similarity with valid X and y
        """
        oz = JointPlot(hist=True, columns=None)
        assert oz.fit(self.discrete.X[:, 0], self.discrete.y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_columns_none_x_hist(self):
        """
        When self.columns=None image similarity with valid X, no y
        """
        oz = JointPlot(hist=True, columns=None)
        assert oz.fit(self.discrete.X) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_columns_single_int_index_numpy_hist(self):
        """
        When self.columns=int image similarity on numpy dataset
        """
        oz = JointPlot(columns=1, hist=True)
        assert oz.fit(self.continuous.X, self.continuous.y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        # Python 3.6 Travis ImageComparisonFailure: images not close (RMS 0.837)
        self.assert_images_similar(oz, tol=1.0)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_columns_single_str_index_pandas_hist(self):
        """
        When self.columns=str image similarity on pandas dataset
        """
        oz = JointPlot(columns="a", hist=True)
        X = pd.DataFrame(self.continuous.X, columns=["a", "b"])
        y = pd.Series(self.continuous.y)
        assert oz.fit(X, y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        # Python 3.6 Travis ImageComparisonFailure: images not close (RMS 1.109)
        self.assert_images_similar(oz, tol=1.5)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_columns_double_int_index_numpy_no_y_hist(self):
        """
        When self.columns=[int, int] image similarity on numpy dataset no y
        """
        oz = JointPlot(columns=[0, 1], hist=True)
        assert oz.fit(self.discrete.X, y=None) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_columns_double_str_index_pandas_no_y_hist(self):
        """
        When self.columns=[str, str] image similarity on pandas dataset no y
        """
        oz = JointPlot(columns=["a", "b"], hist=True)
        X = pd.DataFrame(self.continuous.X, columns=["a", "b"])
        assert oz.fit(X, y=None) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_columns_double_index_discrete_y_hist(self):
        """
        When self.columns=[str, str] on DataFrame with discrete y
        """
        oz = JointPlot(columns=["a", "b"], hist=True)
        X = pd.DataFrame(self.discrete.X, columns=["a", "b"])
        y = pd.Series(self.discrete.y)
        assert oz.fit(X, y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_columns_double_index_continuous_y_hist(self):
        """
        When self.columns=[str, str] on DataFrame with continuous y
        """
        oz = JointPlot(columns=["a", "b"], hist=True)
        X = pd.DataFrame(self.continuous.X, columns=["a", "b"])
        y = pd.Series(self.continuous.y)
        assert oz.fit(X, y) is oz
        assert hasattr(oz, "corr_")

        oz.finalize()
        # Appveyor and Linux conda failed due to non-text-based differences
        self.assert_images_similar(oz, tol=4.0)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_quick_method(self):
        """
        Test the joint_plot quick method
        """
        oz = joint_plot(self.continuous.X, self.continuous.y, columns=1, show=False)
        assert isinstance(oz, JointPlot)
        assert hasattr(oz, "corr_")

        # Python 3.6 Travis ImageComparisonFailure: images not close (RMS 0.837)
        self.assert_images_similar(oz, tol=1.0)



# tests.test_style.test_colors
# Tests for the color utilities and helpers module
#
# Author:   Benjamin Bengfort <bbengfort@districtdatalabs.com>
# Created:  Thu Oct 06 09:30:49 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_colors.py [c6aff34] benjamin@bengfort.com $

"""
Tests for the color utilities and helper functions
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest

from matplotlib import cm
from cycler import Cycler

from sklearn.cluster import KMeans
from sklearn.datasets import make_blobs

from yellowbrick.style.colors import *
from yellowbrick.style.palettes import ColorPalette, PALETTES
from yellowbrick.cluster.silhouette import SilhouetteVisualizer

from tests.base import VisualTestCase


##########################################################################
## Color Tests
##########################################################################


class TestGetColorCycle(VisualTestCase):
    """
    Test get_color_cycle helper function
    """

    def test_cycle_depends_on_palette(self):
        """
        Ensure the color cycle depends on the palette
        """
        c = get_color_cycle()
        assert len(c) == 6

        with ColorPalette("paired"):
            c = get_color_cycle()
            assert len(c) == 12

        c = get_color_cycle()
        assert len(c) == 6

    @pytest.mark.filterwarnings()
    def test_color_cycle_matches_prop_cycle(self):
        """
        Test get color cycle returns colors from axes.prop_cycle
        """
        colors = get_color_cycle()
        cycle = mpl.rcParams["axes.prop_cycle"]

        # Ensure the cycle is in fact a cycle
        assert isinstance(cycle, Cycler)

        # Ensure that colors is actually a list (might change in the future)
        assert isinstance(colors, list)

        # Ensure the cycler and the colors have the same length
        cycle = list(cycle)
        assert len(colors) == len(cycle)

        # Ensure the colors and cycle match
        for color, cycle_color in zip(colors, cycle):
            assert color == cycle_color["color"]


class TestResolveColors(VisualTestCase):
    """
    Test resolve_colors helper function
    """

    def test_resolve_colors_default(self):
        """
        Provides reasonable defaults provided no arguments
        """
        colors = resolve_colors()
        assert colors == get_color_cycle()

    def test_resolve_colors_default_truncate(self):
        """
        Truncates default colors when n_colors is smaller than palette
        """
        assert len(get_color_cycle()) > 3
        assert len(resolve_colors(3)) == 3

    def test_resolve_colors_default_multiply(self):
        """
        Multiplies default colors when n_colors is larger than palette
        """
        assert len(get_color_cycle()) < 18
        assert len(resolve_colors(18)) == 18

    def test_warning_on_colormap_and_colors_args(self):
        """
        Warns when both colormap and colors is used, colors is default
        """
        with pytest.warns(Warning, match="both colormap and colors specified"):
            colors = resolve_colors(colormap="RdBu", colors=["r", "g", "b"])
            assert colors == ["r", "g", "b"]

    def test_colormap_invalid(self):
        """
        Exception raised when invalid colormap is supplied
        """
        with pytest.raises(YellowbrickValueError):
            resolve_colors(12, colormap="foo")

    def test_colormap_string(self):
        """
        Check resolve colors works when a colormap string is passed
        """
        cases = (
            (
                {"n_colors": 6, "colormap": "RdBu"},
                [
                    (0.403921568627451, 0.0, 0.12156862745098039, 1.0),
                    (0.8392156862745098, 0.3764705882352941, 0.30196078431372547, 1.0),
                    (0.9921568627450981, 0.8588235294117647, 0.7803921568627451, 1.0),
                    (0.8196078431372551, 0.8980392156862746, 0.9411764705882353, 1.0),
                    (0.2627450980392157, 0.5764705882352941, 0.7647058823529411, 1.0),
                    (0.0196078431372549, 0.18823529411764706, 0.3803921568627451, 1.0),
                ],
            ),
            (
                {"n_colors": 18, "colormap": "viridis"},
                [
                    (0.267004, 0.004874, 0.329415, 1.0),
                    (0.281924, 0.089666, 0.412415, 1.0),
                    (0.280255, 0.165693, 0.476498, 1.0),
                    (0.263663, 0.237631, 0.518762, 1.0),
                    (0.237441, 0.305202, 0.541921, 1.0),
                    (0.208623, 0.367752, 0.552675, 1.0),
                    (0.182256, 0.426184, 0.55712, 1.0),
                    (0.159194, 0.482237, 0.558073, 1.0),
                    (0.13777, 0.537492, 0.554906, 1.0),
                    (0.121148, 0.592739, 0.544641, 1.0),
                    (0.128087, 0.647749, 0.523491, 1.0),
                    (0.180653, 0.701402, 0.488189, 1.0),
                    (0.274149, 0.751988, 0.436601, 1.0),
                    (0.395174, 0.797475, 0.367757, 1.0),
                    (0.535621, 0.835785, 0.281908, 1.0),
                    (0.688944, 0.865448, 0.182725, 1.0),
                    (0.845561, 0.887322, 0.099702, 1.0),
                    (0.993248, 0.906157, 0.143936, 1.0),
                ],
            ),
            (
                {"n_colors": 9, "colormap": "Set1"},
                [
                    (0.8941176470588236, 0.10196078431372549, 0.10980392156862745, 1.0),
                    (0.21568627450980393, 0.49411764705882355, 0.7215686274509804, 1.0),
                    (0.30196078431372547, 0.6862745098039216, 0.2901960784313726, 1.0),
                    (0.596078431372549, 0.3058823529411765, 0.6392156862745098, 1.0),
                    (1.0, 0.4980392156862745, 0.0, 1.0),
                    (1.0, 1.0, 0.2, 1.0),
                    (0.6509803921568628, 0.33725490196078434, 0.1568627450980392, 1.0),
                    (0.9686274509803922, 0.5058823529411764, 0.7490196078431373, 1.0),
                    (0.6, 0.6, 0.6, 1.0),
                ],
            ),
        )

        for kwds, expected in cases:
            colors = resolve_colors(**kwds)
            assert isinstance(colors, list)
            assert colors == expected

    def test_colormap_string_default_length(self):
        """
        Check colormap when n_colors is not specified
        """
        n_colors = len(get_color_cycle())
        assert len(resolve_colors(colormap="autumn")) == n_colors

    def test_colormap_cmap(self):
        """
        Assert that supplying a maptlotlib.cm as colormap works
        """
        cmap = cm.get_cmap("nipy_spectral")
        colors = resolve_colors(4, colormap=cmap)
        assert colors == [
            (0.0, 0.0, 0.0, 1.0),
            (0.0, 0.6444666666666666, 0.7333666666666667, 1.0),
            (0.7999666666666666, 0.9777666666666667, 0.0, 1.0),
            (0.8, 0.8, 0.8, 1.0),
        ]

    def test_colormap_palette_mpl(self):
        """
        Assert that supplying a maptlotlib palette as colormap works
        """
        colormap = cm.get_cmap("nipy_spectral")
        colors = resolve_colors(colormap=colormap)
        assert colors == [
            (0.0, 0.0, 0.0, 1.0),
            (0.0, 0.0, 0.8667, 1.0),
            (0.0, 0.6667, 0.5333, 1.0),
            (0.0, 1.0, 0.0, 1.0),
            (1.0, 0.6, 0.0, 1.0),
            (0.8, 0.8, 0.8, 1.0),
        ]

    def test_integrated_yb_colormap(self):
        """
        Assert silhouette plot colormap can be set with a yellowbrick palette
        """
        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=8, shuffle=False, random_state=0
        )
        visualizer = SilhouetteVisualizer(
            KMeans(random_state=0), colormap="neural_paint"
        )
        visualizer.fit(X)
        visualizer.finalize()

        tol = (
            3.2 if sys.platform == "win32" else 0.01
        )  # Fails on AppVeyor with RMS 3.143
        self.assert_images_similar(visualizer, remove_legend=True, tol=tol)

    def test_colormap_palette_yb(self):
        """
        Assert that supplying a yellowbrick palette as colormap works
        """
        colormap = ColorPalette("neural_paint")
        assert resolve_colors(colormap=colormap) == [
            (0.08627450980392157, 0.44313725490196076, 0.5725490196078431),
            (0.43137254901960786, 0.4588235294117647, 0.2823529411764706),
            (0.7725490196078432, 0.6352941176470588, 0.6705882352941176),
            (0.0, 0.8, 1.0),
            (0.8705882352941177, 0.47058823529411764, 0.6823529411764706),
            (1.0, 0.8, 0.6),
            (0.23921568627450981, 0.24705882352941178, 0.25882352941176473),
            (1.0, 1.0, 0.8),
        ]

    def test_colormap_cmap_with_colors(self):
        """
        Assert that colors overrides a mpl colormap if both are provided
        """
        colormap = cm.get_cmap("nipy_spectral")
        overriding_colors = [
            (0.0, 0.0, 0.0, 1.0),
            (0.0, 0.6444666666666666, 0.7333666666666667, 1.0),
            (0.7999666666666666, 0.9777666666666667, 0.0, 1.0),
            (0.8, 0.8, 0.8, 1.0),
        ]
        with pytest.warns(Warning, match="both colormap and colors specified"):
            colors = resolve_colors(colormap=colormap, colors=overriding_colors)
            assert colors == overriding_colors

    def test_colormap_palette_yb_colors(self):
        """
        Assert that colors overrides a yellowbrick colormap if both are provided
        """
        colormap = ColorPalette("neural_paint")
        overriding_colors = [
            (0.0, 0.0, 0.0, 1.0),
            (0.0, 0.6444666666666666, 0.7333666666666667, 1.0),
            (0.7999666666666666, 0.9777666666666667, 0.0, 1.0),
            (0.8, 0.8, 0.8, 1.0),
        ]
        with pytest.warns(Warning, match="both colormap and colors specified"):
            colors = resolve_colors(colormap=colormap, colors=overriding_colors)
            assert colors == overriding_colors

    def test_colormap_invalid_type(self):
        """
        Exception raised when invalid colormap type is supplied
        """
        with pytest.raises(YellowbrickValueError):
            resolve_colors(colormap=lambda x: x + 1)

    def test_colors(self):
        """
        Test passing in a list of colors
        """
        c = PALETTES["flatui"]
        assert resolve_colors(colors=c) == c

    def test_colors_truncate(self):
        """
        Test passing in a list of colors with n_colors truncate
        """
        c = PALETTES["flatui"]

        assert len(c) > 3
        assert len(resolve_colors(n_colors=3, colors=c)) == 3

    def test_colors_multiply(self):
        """
        Test passing in a list of colors with n_colors multiply
        """
        c = PALETTES["flatui"]

        assert len(c) < 12
        assert len(resolve_colors(n_colors=12, colors=c)) == 12



# tests.images
# Helper utility to manage baseline images for image comparisons.
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Fri Mar 02 21:51:56 2018 -0500
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: images.py [ab37b18] benjamin@bengfort.com $

"""
Helper utility to manage baseline images for image comparisons. Usage:

    $ python -m tests.images --help

The utility uses argparse to manage command line input for specific baseline
image movement or clearing.
"""

##########################################################################
## Imports
##########################################################################

import os
import glob
import shutil
import argparse


BASE = os.path.dirname(__file__)
BASELINE = os.path.join(BASE, "baseline_images")
ACTUAL = os.path.join(BASE, "actual_images")


##########################################################################
## Helper Methods
##########################################################################


def relpath(path):
    """
    Compute the path relative to the test directory.
    """
    path = path.rstrip(".py")
    path = os.path.relpath(path, start=BASE)
    if path.startswith("..") or path.startswith(os.path.sep):
        raise ValueError("path must include test directory")
    return path


def validate(path):
    """
    Validates a test directory path by checking if any actual images exist
    and creating the mirror baseline directory if required. Returns bool to
    show if work should be done on the path or raises an exception if invalid.
    """
    actual_path = os.path.join(ACTUAL, path)
    if not os.path.exists(actual_path):
        return False

    if not os.path.isdir(actual_path):
        raise ValueError("{} is not a directory".format(actual_path))

    baseline_path = os.path.join(BASELINE, path)
    if not os.path.exists(baseline_path):
        os.makedirs(baseline_path)

    return True


def clear(path):
    """
    Clear .png files in the actual and baseline directories
    """
    for root in (BASELINE, ACTUAL):
        for img in glob.glob(os.path.join(root, path, "*.png")):
            os.remove(img)
            print("removed {}".format(os.path.relpath(img)))


def list_images(path):
    """
    List the associated files for the given test path
    """
    # Get contents of directories
    bases = set(glob.glob(os.path.join(BASELINE, path, "*.png")))
    actual = set(glob.glob(os.path.join(ACTUAL, path, "*.png")))
    diffs = set(glob.glob(os.path.join(ACTUAL, path, "*-failed-diff.png")))

    # Dedupe actual and diffs
    actual -= diffs

    # Get test file names without parent dirs
    names = set(map(os.path.basename, bases)) | set(map(os.path.basename, actual))

    # Build listing
    output = []

    for name in names:
        bname, bext = os.path.splitext(name)

        output.append(bname)
        output.append("-" * len(bname))

        # Handle base path
        base_path = os.path.join(BASELINE, path, name)
        if base_path in bases:
            output.append("  - {}".format(os.path.relpath(base_path)))
        else:
            output.append("  - no baseline image")

        # Handle actual path
        actual_path = os.path.join(ACTUAL, path, name)
        if actual_path in actual:
            output.append("  - {}".format(os.path.relpath(actual_path)))
        else:
            output.append("  - no actual image")

        # Handle diff path
        diff_path = os.path.join(ACTUAL, path, "{}-failed-diff{}".format(bname, bext))
        if diff_path in diffs:
            output.append("  - {}".format(os.path.relpath(diff_path)))

        # Add breathing room
        output.append("")

    return "\n".join(output)


def sync(path):
    """
    Move all non-diff images from actual to baseline
    """
    for fname in os.listdir(os.path.join(ACTUAL, path)):
        if fname.endswith("-diff.png"):
            continue

        if fname.endswith(".png"):
            src = os.path.join(ACTUAL, path, fname)
            dst = os.path.join(BASELINE, path, fname)
            shutil.copy2(src, dst)
            print("synced {}".format(os.path.relpath(src, start=ACTUAL)))


##########################################################################
## Main Method
##########################################################################


def main(args):
    # Get directories relative to test dir
    test_dirs = list(map(relpath, args.test_dirs))

    # Validate directories and filter empty ones
    test_dirs = filter(validate, test_dirs)

    # If list, simply list what is available and exit
    if args.list:
        output = [list_images(path) for path in test_dirs]
        print("\n".join(output))
        return

    # If clear, clear the baseline and actual directories
    if args.clear:
        for path in test_dirs:
            clear(path)
        return

    # Move images that aren't diffs from actual to baseline
    for path in test_dirs:
        sync(path)


if __name__ == "__main__":

    args = {
        ("-C", "--clear"): {
            "action": "store_true",
            "help": "clear actual, diffs, and baseline images for test",
        },
        ("-L", "--list"): {
            "action": "store_true",
            "help": "list images images for tests and exit",
        },
        "test_dirs": {
            "metavar": "DIR",
            "nargs": "+",
            "help": "directories to move images from actual to baseline",
        },
    }

    # Create the parser and add the arguments
    parser = argparse.ArgumentParser(
        description="utility to manage baseline images for comparisons",
        epilog="report any issues on GitHub",
    )

    for pargs, kwargs in args.items():
        if isinstance(pargs, str):
            pargs = (pargs,)
        parser.add_argument(*pargs, **kwargs)

    args = parser.parse_args()
    try:
        main(args)
    except Exception as e:
        parser.error(str(e))



# tests.test_text.test_correlation
# Tests for the dispersion plot visualization
#
# Author:   Patrick Deziel
# Created:  Tue May 3 16:18:21 2022 -0500
#
# Copyright (C) 2022 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_correlation.py [0a7d2fe] deziel.patrick@gmail.com $

"""
Tests for the word correlation plot text visualization
"""

##########################################################################
## Imports
##########################################################################

import pytest
import matplotlib.pyplot as plt

from tests.base import VisualTestCase
from yellowbrick.datasets import load_hobbies
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.text.correlation import WordCorrelationPlot, word_correlation

##########################################################################
## Data
##########################################################################

corpus = load_hobbies()

##########################################################################
## WordCorrelationPlot Tests
##########################################################################

class TestWordCorrelationPlot(VisualTestCase):
    def test_quick_method(self):
        """
        Assert no errors occur when using the quick method.
        """
        _, ax = plt.subplots()
        words = ["Game", "player", "score", "oil"]

        viz = word_correlation(words, corpus.data, ax=ax, show=False)

        self.assert_images_similar(viz, tol=25)

    def test_word_correlation_plot(self):
        """
        Assert no errors are raised during normal execution.
        """
        words = ["Game", "player", "score", "oil", "Man"]

        viz = WordCorrelationPlot(words)
        assert viz.fit(corpus.data) is viz, "fit method should return self"

        self.assert_images_similar(viz, tol=25)

    def test_word_correlation_generator(self):
        """
        Assert no errors are raised when the corpus is a generator.
        """
        words = ["Game", "player", "score", "oil", "Man", "woman"]

        viz = WordCorrelationPlot(words)

        def stream_corpus(data):
            for doc in data:
                yield doc
        viz.fit(stream_corpus(corpus.data))

        self.assert_images_similar(viz, tol=25)

    def test_word_correlation_ignore_case(self):
        """
        Assert no errors are raised when ignore_case is True.
        """
        words = ["Game", "player", "score", "oil", "game"]

        viz = WordCorrelationPlot(words, ignore_case=True)
        viz.fit(corpus.data)

        self.assert_images_similar(viz, tol=25)

    def test_word_correlation_ngrams(self):
        """
        Assert no errors are raised when multiple-word terms are provided.
        """
        words = ["Tatsumi Kimishima", "Nintendo", "game", "man", "play"]

        viz = WordCorrelationPlot(words)
        viz.fit(corpus.data)

        self.assert_images_similar(viz, tol=25)

    def test_word_correlation_no_words(self):
        """
        Assert that an error is raised when no words are provided.
        """
        with pytest.raises(YellowbrickValueError):
            WordCorrelationPlot([])

        with pytest.raises(YellowbrickValueError):
            WordCorrelationPlot([""])

        with pytest.raises(YellowbrickValueError):
            WordCorrelationPlot([" ", "\t", "\n"])

    def test_word_correlation_missing_words(self):
        """
        Assert that an error is raised on fit() when a word does not exist in the
        corpus.
        """
        words = ["Game", "player", "score", "oil", "NotACorpusWord"]
        
        viz = WordCorrelationPlot(words)
        with pytest.raises(YellowbrickValueError):
            viz.fit(corpus.data)


# tests.test_contrib.test_wrapper
# Tests third-party estimator wrapper
#
# Author:   Benjamin Bengfort
# Created:  Fri Oct 02 17:27:50 2020 -0400
#
# Copyright (C) 2020 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_wrapper.py [] benjamin@bengfort.com $

"""
Tests third-party estimator wrapper
"""

##########################################################################
## Imports
##########################################################################

import pytest

from yellowbrick.contrib.wrapper import *
from yellowbrick.regressor import residuals_plot
from yellowbrick.classifier import classification_report
from yellowbrick.exceptions import YellowbrickAttributeError
from yellowbrick.utils.helpers import get_model_name, is_fitted
from yellowbrick.utils.types import is_estimator, is_probabilistic
from yellowbrick.utils.types import is_classifier, is_clusterer, is_regressor

from sklearn.base import is_regressor as sk_is_regressor
from sklearn.base import is_classifier as sk_is_classifier
from sklearn.model_selection import train_test_split as tts
from sklearn.datasets import make_classification, make_regression
from sklearn.base import is_outlier_detector as sk_is_outlier_detector


try:
    import xgboost as xgb
except ImportError:
    xgb = None

try:
    import cudf
    from cuml.ensemble import RandomForestClassifier as curfc
except ImportError:
    curfc = None

try:
    import catboost
except ImportError:
    catboost = None


##########################################################################
## Mocks and Fixtures
##########################################################################

class ThirdPartyEstimator(object):

    def __init__(self, **params):
        for attr, param in params.items():
            setattr(self, attr, param)

    def fit(self, X, y=None):
        return 42

    def predict_proba(self, X, y=None):
        return 24


##########################################################################
## Test Suite
##########################################################################

class TestContribWrapper(object):
    """
    Third party ContribEstimator wrapper
    """

    def test_wrapped_estimator(self):
        """
        Check that the contrib wrapper passes through correctly
        """
        tpe = ContribEstimator(ThirdPartyEstimator(foo="bar"), "foo")
        assert tpe.fit([1,2,3]) == 42
        assert tpe._estimator_type == "foo"

    def test_attribute_error(self):
        """
        Assert a correct exception is raised on failed access
        """
        tpe = ContribEstimator(ThirdPartyEstimator())
        with pytest.raises(YellowbrickAttributeError):
            tpe.foo

    def test_get_model_name(self):
        """
        ContribWrapper should return the underlying model name
        """
        tpe = ContribEstimator(ThirdPartyEstimator())
        assert get_model_name(tpe) == "ThirdPartyEstimator"

    def test_wraps_is_estimator(self):
        """
        Assert a wrapped estimator passes is_estimator check
        """
        tpe = wrap(ThirdPartyEstimator())
        assert is_estimator(tpe)

    def test_wraps_is_classifier(self):
        """
        Assert a wrapped estimator passes is_classifier check
        """
        tpe = classifier(ThirdPartyEstimator())
        assert is_classifier(tpe)
        assert sk_is_classifier(tpe)

    def test_wraps_is_regressor(self):
        """
        Assert a wrapped estimator passes is_regressor check
        """
        tpe = regressor(ThirdPartyEstimator())
        assert is_regressor(tpe)
        assert sk_is_regressor(tpe)

    def test_wraps_is_clusterer(self):
        """
        Assert a wrapped estimator passes is_clusterer check
        """
        tpe = clusterer(ThirdPartyEstimator())
        assert is_clusterer(tpe)

    def test_wraps_is_outlier_detector(self):
        """
        Assert a wrapped estimator passes is_outlier_detector check
        """
        tpe = wrap(ThirdPartyEstimator(), OUTLIER_DETECTOR)
        assert sk_is_outlier_detector(tpe)

    def test_wraps_is_probabilistic(self):
        """
        Assert a wrapped estimator passes is_probabilistic check
        """
        tpe = wrap(ThirdPartyEstimator())
        assert is_probabilistic(tpe)


##########################################################################
## Test Non-sklearn Estimators
##########################################################################

class TestNonSklearnEstimators(object):
    """
    Check various non-sklearn estimators to see if the wrapper works for them
    """

    @pytest.mark.skipif(xgb is None, reason="requires xgboost")
    def test_xgboost_regressor(self):
        """
        Validate xgboost regressor with wrapper
        """
        X, y = make_regression(
            n_samples=500, n_features=22, n_informative=8, random_state=8311982
        )
        X_train, X_test, y_train, y_test = tts(X, y)

        model = regressor(xgb.XGBRFRegressor())
        oz = residuals_plot(model, X_train, y_train, X_test, y_test, show=False)
        assert is_fitted(oz)


    @pytest.mark.skipif(xgb is None, reason="requires xgboost")
    def test_xgboost_regressor_unwrapped(self):
        """
        Validate xgboost regressor without wrapper
        """
        X, y = make_regression(
            n_samples=500, n_features=22, n_informative=8, random_state=8311982
        )
        X_train, X_test, y_train, y_test = tts(X, y)

        model = xgb.XGBRFRegressor()
        oz = residuals_plot(model, X_train, y_train, X_test, y_test, show=False)
        assert is_fitted(oz)

    @pytest.mark.skipif(curfc is None, reason="requires cuML")
    def test_cuml_classifier(self):
        """
        Validate cuML classifier with wrapper
        """
        # NOTE: this is currently untested as I wasn't able to install cuML
        X, y = make_classification(
            n_samples=400, n_features=10, n_informative=2, n_redundant=3,
            n_classes=2, n_clusters_per_class=2, random_state=8311982
        )
        X_train, X_test, y_train, y_test = tts(X, y)

        # Convert to cudf dataframes
        X_train = cudf.DataFrame(X_train)
        y_train = cudf.Series(y_train)
        X_test = cudf.DataFrame(X_test)
        y_test = cudf.Series(y_test)

        model = classifier(curfc(n_estimators=40, max_depth=8, max_features=1))
        oz = classification_report(model, X_train, y_train, X_test, y_test, show=False)
        assert is_fitted(oz)

    @pytest.mark.skipif(catboost is None, reason="requires CatBoost")
    def test_catboost_classifier(self):
        """
        Validate CatBoost classifier with wrapper
        """
        X, y = make_classification(
            n_samples=400, n_features=10, n_informative=2, n_redundant=3,
            n_classes=2, n_clusters_per_class=2, random_state=8311982
        )
        X_train, X_test, y_train, y_test = tts(X, y)

        model = classifier(catboost.CatBoostClassifier(
            iterations=2, depth=2, learning_rate=1, loss_function='Logloss'
        ))

        # For some reason, this works if you call fit directly and pass is_fitted to
        # the visualizer, but does not work if you rely on the visualizer to fit the
        # model on the data. I can't tell if this is a catboost or Yellowbrick issue.
        model.fit(X_train, y_train)

        oz = classification_report(
            model, X_train, y_train, X_test, y_test, is_fitted=True, show=False
        )
        assert is_fitted(oz)



# tests.test_cluster.test_base
# Test the cluster base visualizers.
#
# Author:   Rebecca Bilbro
# Created:  Thu Mar 23 17:38:42 2017 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_base.py [241edca] benjamin@bengfort.com $

"""
Test the cluster base visualizers.
"""

##########################################################################
## Imports
##########################################################################

import pytest

from yellowbrick.exceptions import YellowbrickTypeError
from yellowbrick.cluster.base import ClusteringScoreVisualizer

from sklearn.svm import SVC, SVR
from sklearn.ensemble import RandomForestClassifier
from sklearn.cluster import MeanShift, DBSCAN, Birch
from sklearn.linear_model import Ridge, RidgeCV, LinearRegression
from sklearn.cluster import KMeans, MiniBatchKMeans, AffinityPropagation


##########################################################################
## Clustering Base Test Cases
##########################################################################


class TestClusterBase(object):
    @pytest.mark.parametrize(
        "model", [SVC, SVR, Ridge, RidgeCV, LinearRegression, RandomForestClassifier]
    )
    def test_clusterer_enforcement_raises(self, model):
        """
        Assert that non-cluster models raise a TypeError for cluster visualizers
        """
        with pytest.raises(YellowbrickTypeError):
            ClusteringScoreVisualizer(model())

    @pytest.mark.parametrize(
        "model",
        [KMeans, MiniBatchKMeans, AffinityPropagation, MeanShift, DBSCAN, Birch],
    )
    def test_clusterer_enforcement(self, model):
        """
        Assert that only clustering estimators can be passed to cluster viz
        """
        try:
            ClusteringScoreVisualizer(model())
        except YellowbrickTypeError:
            self.fail("could not pass clustering estimator to visualizer")

    def test_force_estimator(self):
        """
        Test that an estimator can be forced through
        """
        with pytest.raises(YellowbrickTypeError):
            ClusteringScoreVisualizer(LinearRegression())

        try:
            ClusteringScoreVisualizer(LinearRegression(), force_model=True)
        except YellowbrickTypeError as e:
            pytest.fail("type error was raised incorrectly: {}".format(e))



# tests.test_contrib.test_statsmodels.test_base
# Tests for the statsmodels estimator wrapper.
#
# Author:  Ian Ozsvald
# Created: Wed Jan 10 12:47:00 2018 -0500
#
# ID: test_base.py [d6ebc39] benjamin@bengfort.com $

"""
Tests for the statsmodels estimator wrapper.
"""

##########################################################################
## Imports
##########################################################################

import pytest
import functools
import numpy as np

from yellowbrick.contrib.statsmodels import StatsModelsWrapper

try:
    import statsmodels.api as sm
except ImportError:
    sm = None


##########################################################################
## Test Cases
##########################################################################


@pytest.mark.skipif(sm is None, reason="test requires statsmodels")
def test_stats_models_wrapper():
    """
    A trivial test of the StatsModelsWrapper
    """
    X = np.array([[1], [2], [3]])
    y = np.array([1.1, 2, 3])

    glm_gaussian = functools.partial(sm.GLM, family=sm.families.Gaussian())
    sm_est = StatsModelsWrapper(glm_gaussian)

    assert sm_est.fit(X, y) is sm_est, "fit did not return self"
    assert sm_est.predict(X).shape == (3,)
    assert 0.0 <= sm_est.score(X, y) <= 1.0



# tests.test_target.test_binning
# Tests for the BalancedBinningReference visualizer
#
# Author:  Juan L. Kehoe
# Author:  Prema Damodaran Roman
# Created: Thu Jul 20 10:21:49 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_binning.py

##########################################################################
# Imports
##########################################################################
import pytest

from yellowbrick.target.binning import *
from yellowbrick.datasets import load_occupancy

from tests.base import VisualTestCase

try:
    import pandas as pd
except ImportError:
    pd = None

##########################################################################
# BalancedBinningReference Tests
##########################################################################


class TestBalancedBinningReference(VisualTestCase):
    """
    Test the BalancedBinningReference visualizer
    """

    def test_numpy_bins(self):
        """
        Test Histogram on a NumPy array
        """
        # Load the data from the fixture
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()

        visualizer = BalancedBinningReference()
        visualizer.fit(y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.5)

    @pytest.mark.skipif(pd is None, reason="pandas is required")
    def test_pandas_bins(self):
        """
        Test Histogram on a Pandas Dataframe
        """
        # Load the data from the fixture
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()

        visualizer = BalancedBinningReference()
        visualizer.fit(y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.5)

    def test_quick_method(self):
        """
        Test the quick method with producing a valid visualization
        """
        data = load_occupancy(return_dataset=True)
        _, y = data.to_numpy()

        visualizer = balanced_binning_reference(y, show=False)

        assert isinstance(visualizer, BalancedBinningReference)
        self.assert_images_similar(visualizer, tol=0.5)



# tests.test_contrib.test_scatter
# Test the ScatterViz feature analysis visualizers
#
# Author:   Nathan Danielsen
# Created:  Fri Feb 26 19:40:00 2017 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_scatter.py [a89633e] benjamin@bengfort.com $
"""
Test the ScatterViz feature analysis visualizers
"""

##########################################################################
# Imports
##########################################################################

import pytest
import numpy as np

from unittest import mock
from tests.base import VisualTestCase
from yellowbrick.contrib.scatter import *
from yellowbrick.datasets import load_occupancy
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.style import palettes

try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
# ScatterViz Base Tests
##########################################################################


@pytest.mark.filterwarnings("ignore")
class TestScatterViz(VisualTestCase):
    """
    Test ScatterViz
    """

    # yapf: disable
    X = np.array([
        [2.318, 2.727, 4.260, 7.212, 4.792],
        [2.315, 2.726, 4.295, 7.140, 4.783, ],
        [2.315, 2.724, 4.260, 7.135, 4.779, ],
        [2.110, 3.609, 4.330, 7.985, 5.595, ],
        [2.110, 3.626, 4.330, 8.203, 5.621, ],
        [2.110, 3.620, 4.470, 8.210, 5.612, ]
    ])
    # yapf: enable
    y = np.array([1, 0, 1, 0, 1, 0])

    def test_init_alias(self):
        """
        Test alias for ScatterViz
        """
        features = ["temperature", "relative humidity"]
        visualizer = ScatterVisualizer(features=features, markers=["*"])
        assert visualizer.markers is not None

    def test_scatter(self):
        """
        Assert no errors occur during scatter visualizer integration
        """
        X_two_cols = self.X[:, :2]
        features = ["temperature", "relative humidity"]
        visualizer = ScatterViz(features=features)
        visualizer.fit_transform(X_two_cols, self.y)

    def test_color_builds(self):
        """
        Assert no errors occur during scatter visualizer integration
        """
        colors = palettes.PALETTES["pastel"]
        X_two_cols = self.X[:, :2]
        features = ["temperature", "relative humidity"]
        visualizer = ScatterViz(features=features, color=colors)
        visualizer.fit_transform(X_two_cols, self.y)

    def test_scatter_no_features(self):
        """
        Assert no errors during scatter visualizer integration - no features
        """
        X_two_cols = self.X[:, :2]
        visualizer = ScatterViz()
        visualizer.fit_transform_show(X_two_cols, self.y)
        assert visualizer.features_ == ["Feature One", "Feature Two"]

    def test_scatter_only_two_features_allowed_init(self):
        """
        Assert that only two features are allowed for scatter visualizer init
        """
        features = ["temperature", "relative humidity", "light"]

        with pytest.raises(YellowbrickValueError):
            ScatterViz(features=features)

    def test_scatter_xy_and_features_raise_error(self):
        """
        Assert that x,y and features will raise scatterviz error
        """
        features = ["temperature", "relative humidity", "light"]

        with pytest.raises(YellowbrickValueError):
            ScatterViz(features=features, x="one", y="two")

    def test_scatter_xy_changes_to_features(self):
        """
        Assert that x,y with no features will not raise scatterviz error
        """
        visualizer = ScatterViz(x="one", y="two")
        assert visualizer.features == ["one", "two"]

    def test_scatter_requires_two_features_in_numpy_matrix(self):
        """
        Assert only two features allowed for scatter visualizer if not in init
        """
        visualizer = ScatterViz()
        with pytest.raises(YellowbrickValueError, match="only accepts two features"):
            visualizer.fit_transform(self.X, self.y)

    def test_integrated_scatter(self):
        """
        Test scatter on the real, occupancy data set
        """
        # Load the data from the fixture
        X, y = load_occupancy(return_dataset=True).to_numpy()

        # Test the visualizer
        features = ["temperature", "relative humidity"]
        visualizer = ScatterViz(features=features)
        visualizer.fit_transform_show(X[:, :2], y)

    def test_alpha_param(self):
        """
        Test that the user can supply an alpha param on instantiation
        """
        # Instantiate a scatter plot and provide a custom alpha
        visualizer = ScatterVisualizer(alpha=0.7, features=["a", "b"])

        # Test param gets set correctly
        assert visualizer.alpha == 0.7

        # Mock ax and fit the visualizer
        visualizer.ax = mock.MagicMock(autospec=True)
        visualizer.fit(self.X[:, :2], self.y)

        # Test that alpha was passed to the scatter plot
        _, scatter_kwargs = visualizer.ax.scatter.call_args
        assert "alpha" in scatter_kwargs
        assert scatter_kwargs["alpha"] == 0.7

    def test_scatter_quick_method(self):
        """
        Test scatter quick method on the real, occupancy data set
        """
        # Load the data from the fixture
        X, y = load_occupancy(return_dataset=True).to_numpy()

        # Test the visualizer
        features = ["temperature", "relative humidity"]
        viz = scatterviz(X[:, :2], y=y, ax=None, features=features)

        # test that is returns a matplotlib obj with axes
        assert isinstance(viz, ScatterVisualizer)

    @pytest.mark.skipif(pd is None, reason="pandas is required for this test")
    def test_integrated_scatter_with_pandas(self):
        """
        Test scatterviz on the real, occupancy data set with pandas
        """
        # Load the data from the fixture
        # Load the data from the fixture
        X, y = load_occupancy(return_dataset=True).to_pandas()

        # Test the visualizer
        features = ["temperature", "relative humidity"]
        visualizer = ScatterViz(features=features)
        visualizer.fit_transform_show(X, y)

    @pytest.mark.xfail(reason="numpy structured arrays have changed since v1.14")
    def test_integrated_scatter_numpy_named_arrays(self):
        """
        Test scatterviz on numpy named arrays
        """
        dt = np.dtype(
            {
                "names": ["one", "two", "three", "four", "five"],
                "formats": [np.float64, np.float64, np.float64, np.float64, np.float64],
            }
        )

        X_named = self.X.astype(dt, casting="unsafe")
        visualizer = ScatterViz(features=["one", "two"])
        visualizer.fit_transform_show(X_named, self.y)
        assert visualizer.features_ == ["one", "two"]

    def test_integrated_scatter_numpy_arrays_no_names(self):
        """
        Test scaterviz on regular numpy arrays
        """
        visualizer = ScatterViz(features=[1, 2])
        visualizer.fit_transform_show(self.X, self.y)
        assert visualizer.features_ == [1, 2]

    def test_scatter_image(self):
        """
        Test the scatterviz image similarity
        """
        # self.setUp_ImageTest()

        X_two_cols = self.X[:, :2]
        features = ["temperature", "relative humidity"]
        visualizer = ScatterViz(features=features)
        visualizer.fit(X_two_cols, self.y)
        visualizer.draw(X_two_cols, self.y)

        self.assert_images_similar(visualizer)



# tests.test_features.test_rankd
# Test the rankd feature analysis visualizers
#
# Author:   Benjamin Bengfort
# Created:  Fri Oct 07 12:19:19 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_rankd.py [7b4350a] nathan.danielsen@gmail.com $

"""
Test the Rankd feature analysis visualizers
"""

##########################################################################
## Imports
##########################################################################

import pytest
import numpy as np
import numpy.testing as npt

from yellowbrick.features.rankd import RankDBase
from yellowbrick.features.rankd import kendalltau
from yellowbrick.features.rankd import Rank1D, rank1d
from yellowbrick.features.rankd import Rank2D, rank2d
from yellowbrick.exceptions import YellowbrickValueError
from tests.base import IS_WINDOWS_OR_CONDA, VisualTestCase
from yellowbrick.datasets import load_occupancy, load_credit, load_energy

try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
## Kendall-Tau Tests
##########################################################################


class TestKendallTau(object):
    """
    Test the Kendall-Tau correlation metric
    """

    def test_kendalltau(self):
        """
        Test results returned match expectations
        """
        X, _ = load_energy(return_dataset=True).to_numpy()

        expected = np.array(
            [
                [1.0, -1.0, -0.2724275, -0.7361443, 0.7385489, 0.0, 0.0, 0.0],
                [-1.0, 1.0, 0.2724275, 0.7361443, -0.7385489, 0.0, 0.0, 0.0],
                [-0.2724275, 0.2724275, 1.0, -0.15192004, 0.19528337, 0.0, 0.0, 0.0],
                [-0.73614431, 0.73614431, -0.15192004, 1.0, -0.87518995, 0.0, 0.0, 0.0],
                [0.73854895, -0.73854895, 0.19528337, -0.87518995, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.15430335],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.15430335, 1.0],
            ]
        )
        actual = kendalltau(X)
        npt.assert_almost_equal(expected, actual)

    def test_kendalltau_shape(self):
        """
        Assert that a square correlation matrix is returned
        """
        X, _ = load_energy(return_dataset=True).to_numpy()
        corr = kendalltau(X)
        assert corr.shape[0] == corr.shape[1]

        for (i, j), val in np.ndenumerate(corr):
            assert corr[j][i] == pytest.approx(val)

    def test_kendalltau_1D(self):
        """
        Assert that a 2D matrix is required as input
        """
        with pytest.raises(IndexError, match="tuple index out of range"):
            X = 0.1 * np.arange(10)
            kendalltau(X)


##########################################################################
## RankDBase Tests
##########################################################################


class TestRankDBase(VisualTestCase):
    """
    Test the RankDBase Visualizer
    """

    def test_rankdbase_unknown_algorithm(self):
        """
        Assert that unknown algorithms raise an exception
        """
        X, _ = load_energy(return_dataset=True).to_numpy()
        with pytest.raises(
            YellowbrickValueError, match=".* is unrecognized ranking method"
        ) as e:
            oz = RankDBase(algorithm="unknown")
            oz.fit_transform(X)
            assert str(e.value) == "'unknown' is unrecognized ranking method"


##########################################################################
## Rank1D Base Tests
##########################################################################


class TestRank1D(VisualTestCase):
    """
    Test the Rank1D visualizer
    """

    def test_rank1d_unknown_algorithm(self):
        """
        Test that an error is raised for Rank1D with an unknown algorithm
        """
        X, _ = load_energy()
        msg = "'oscar' is unrecognized ranking method"
        with pytest.raises(YellowbrickValueError, match=msg):
            Rank1D(algorithm="Oscar").transform(X)

    def test_rank1d_shapiro(self):
        """
        Test Rank1D using shapiro metric
        """
        X, _ = load_energy(return_dataset=True).to_numpy()
        oz = Rank1D(algorithm="shapiro")
        npt.assert_array_equal(oz.fit_transform(X), X)

        # Check Ranking
        expected = np.array(
            [
                0.93340671,
                0.94967198,
                0.92689574,
                0.7459445,
                0.63657606,
                0.85603625,
                0.84349269,
                0.91551381,
            ]
        )

        assert hasattr(oz, "ranks_")
        assert oz.ranks_.shape == (X.shape[1],)
        # Reduce precision to 5 decimals for numerical stability across versions
        npt.assert_array_almost_equal(oz.ranks_, expected, decimal=5)

        # Image similarity comparison
        oz.finalize()
        self.assert_images_similar(oz)

    def test_rank1d_vertical(self):
        """
        Test Rank1D using vertical orientation
        """
        X, _ = load_energy(return_dataset=True).to_numpy()
        oz = Rank1D(orient="v")
        npt.assert_array_equal(oz.fit_transform(X), X)

        # Image similarity comparison
        oz.finalize()
        self.assert_images_similar(oz)

    def test_rank1d_horizontal(self):
        """
        Test Rank1D using horizontal orientation
        """
        X, _ = load_energy(return_dataset=True).to_numpy()
        oz = Rank1D(orient="h")
        npt.assert_array_equal(oz.fit_transform(X), X)

        # Image similarity comparison
        oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.filterwarnings("ignore:p-value")
    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_rank1d_integrated_pandas(self):
        """
        Test Rank1D on occupancy dataset with pandas DataFrame and Series
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()
        features = data.meta["features"]

        assert isinstance(X, pd.DataFrame)
        assert isinstance(y, pd.Series)

        # Test the visualizer
        oz = Rank1D(features=features, show_feature_names=True)
        assert oz.fit(X, y) is oz
        assert oz.transform(X) is X

        # Image similarity testing
        oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.filterwarnings("ignore:p-value")
    def test_rank1d_integrated_numpy(self):
        """
        Test Rank1D on occupancy dataset with default numpy data structures
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()
        features = data.meta["features"]

        assert isinstance(X, np.ndarray)
        assert isinstance(y, np.ndarray)

        # Test the visualizer
        oz = Rank1D(features=features, show_feature_names=True)
        assert oz.fit(X, y) is oz
        assert oz.transform(X) is X

        # Image similarity testing
        oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.filterwarnings("ignore:p-value may not be accurate")
    def test_rank1d_quick_method(self):
        """
        Test Rank1d quick method
        """
        X, y = load_credit()
        viz = rank1d(X, y, show=False)

        assert isinstance(viz, Rank1D)
        self.assert_images_similar(viz, tol=0.1)


##########################################################################
## Rank2D Test Cases
##########################################################################


class TestRank2D(VisualTestCase):
    """
    Test the Rank2D visualizer
    """

    def test_rank2d_unknown_algorithm(self):
        """
        Test that an error is raised for Rank2D with an unknown algorithm
        """
        X, _ = load_energy()
        msg = "'oscar' is unrecognized ranking method"
        with pytest.raises(YellowbrickValueError, match=msg):
            Rank2D(algorithm="Oscar").transform(X)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_rank2d_pearson(self):
        """
        Test Rank2D using pearson metric
        """
        X, _ = load_energy(return_dataset=True).to_numpy()
        oz = Rank2D(algorithm="pearson")
        npt.assert_array_equal(oz.fit_transform(X), X)

        # Check Ranking
        expected = np.array(
            [
                [
                    1.00000000e00,
                    -9.91901462e-01,
                    -2.03781680e-01,
                    -8.68823408e-01,
                    8.27747317e-01,
                    0.00000000e00,
                    1.11706815e-16,
                    -1.12935670e-16,
                ],
                [
                    -9.91901462e-01,
                    1.00000000e00,
                    1.95501633e-01,
                    8.80719517e-01,
                    -8.58147673e-01,
                    0.00000000e00,
                    -2.26567708e-16,
                    -3.55861251e-16,
                ],
                [
                    -2.03781680e-01,
                    1.95501633e-01,
                    1.00000000e00,
                    -2.92316466e-01,
                    2.80975743e-01,
                    0.00000000e00,
                    7.87010445e-18,
                    0.00000000e00,
                ],
                [
                    -8.68823408e-01,
                    8.80719517e-01,
                    -2.92316466e-01,
                    1.00000000e00,
                    -9.72512237e-01,
                    0.00000000e00,
                    -3.27553310e-16,
                    2.20057668e-16,
                ],
                [
                    8.27747317e-01,
                    -8.58147673e-01,
                    2.80975743e-01,
                    -9.72512237e-01,
                    1.00000000e00,
                    0.00000000e00,
                    -1.24094525e-18,
                    0.00000000e00,
                ],
                [
                    0.00000000e00,
                    0.00000000e00,
                    0.00000000e00,
                    0.00000000e00,
                    0.00000000e00,
                    1.00000000e00,
                    -2.42798319e-19,
                    0.00000000e00,
                ],
                [
                    1.11706815e-16,
                    -2.26567708e-16,
                    7.87010445e-18,
                    -3.27553310e-16,
                    -1.24094525e-18,
                    -2.42798319e-19,
                    1.00000000e00,
                    2.12964221e-01,
                ],
                [
                    -1.12935670e-16,
                    -3.55861251e-16,
                    0.00000000e00,
                    2.20057668e-16,
                    0.00000000e00,
                    0.00000000e00,
                    2.12964221e-01,
                    1.00000000e00,
                ],
            ]
        )

        assert hasattr(oz, "ranks_")
        assert oz.ranks_.shape == (X.shape[1], X.shape[1])
        npt.assert_array_almost_equal(oz.ranks_, expected)

        # Image similarity comparision
        oz.finalize()
        # Travis Python 3.6 images not close (RMS 0.112)
        self.assert_images_similar(oz, tol=0.5)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_rank2d_covariance(self):
        """
        Test Rank2D using covariance metric
        """
        X, _ = load_energy(return_dataset=True).to_numpy()
        oz = Rank2D(algorithm="covariance")
        npt.assert_array_equal(oz.fit_transform(X), X)

        # Check Ranking
        expected = np.array(
            [
                [
                    1.11888744e-02,
                    -9.24206867e00,
                    -9.40391134e-01,
                    -4.15083877e00,
                    1.53324641e-01,
                    0.00000000e00,
                    1.57414282e-18,
                    -1.85278419e-17,
                ],
                [
                    -9.24206867e00,
                    7.75916384e03,
                    7.51290743e02,
                    3.50393655e03,
                    -1.32370274e02,
                    0.00000000e00,
                    -2.65874531e-15,
                    -4.86170571e-14,
                ],
                [
                    -9.40391134e-01,
                    7.51290743e02,
                    1.90326988e03,
                    -5.75989570e02,
                    2.14654498e01,
                    0.00000000e00,
                    4.57406096e-17,
                    0.00000000e00,
                ],
                [
                    -4.15083877e00,
                    3.50393655e03,
                    -5.75989570e02,
                    2.03996306e03,
                    -7.69178618e01,
                    0.00000000e00,
                    -1.97089918e-15,
                    1.54151644e-14,
                ],
                [
                    1.53324641e-01,
                    -1.32370274e02,
                    2.14654498e01,
                    -7.69178618e01,
                    3.06649283e00,
                    0.00000000e00,
                    -2.89497529e-19,
                    0.00000000e00,
                ],
                [
                    0.00000000e00,
                    0.00000000e00,
                    0.00000000e00,
                    0.00000000e00,
                    0.00000000e00,
                    1.25162973e00,
                    -3.61871912e-20,
                    0.00000000e00,
                ],
                [
                    1.57414282e-18,
                    -2.65874531e-15,
                    4.57406096e-17,
                    -1.97089918e-15,
                    -2.89497529e-19,
                    -3.61871912e-20,
                    1.77477184e-02,
                    4.40026076e-02,
                ],
                [
                    -1.85278419e-17,
                    -4.86170571e-14,
                    0.00000000e00,
                    1.54151644e-14,
                    0.00000000e00,
                    0.00000000e00,
                    4.40026076e-02,
                    2.40547588e00,
                ],
            ]
        )

        assert hasattr(oz, "ranks_")
        assert oz.ranks_.shape == (X.shape[1], X.shape[1])
        npt.assert_array_almost_equal(oz.ranks_, expected, decimal=5)

        # Image similarity comparision
        oz.finalize()
        self.assert_images_similar(oz, tol=0.1)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_rank2d_spearman(self):
        """
        Test Rank2D using spearman metric
        """
        X, _ = load_energy(return_dataset=True).to_numpy()
        oz = Rank2D(algorithm="spearman")
        npt.assert_array_equal(oz.fit_transform(X), X)

        # Check Ranking
        expected = np.array(
            [
                [1.0, -1.0, -0.25580533, -0.8708862, 0.86904819, 0.0, 0.0, 0.0],
                [-1.0, 1.0, 0.25580533, 0.8708862, -0.86904819, 0.0, 0.0, 0.0],
                [-0.25580533, 0.25580533, 1.0, -0.19345677, 0.22076336, 0.0, 0.0, 0.0],
                [-0.8708862, 0.8708862, -0.19345677, 1.0, -0.93704257, 0.0, 0.0, 0.0],
                [0.86904819, -0.86904819, 0.22076336, -0.93704257, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.18759162],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.18759162, 1.0],
            ]
        )

        assert hasattr(oz, "ranks_")
        assert oz.ranks_.shape == (X.shape[1], X.shape[1])
        npt.assert_array_almost_equal(oz.ranks_, expected)

        # Image similarity comparision
        oz.finalize()
        self.assert_images_similar(oz, tol=0.1)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_rank2d_kendalltau(self):
        """
        Test Rank2D using kendalltau metric
        """
        X, _ = load_energy(return_dataset=True).to_numpy()
        oz = Rank2D(algorithm="kendalltau")
        npt.assert_array_equal(oz.fit_transform(X), X)

        # Check Ranking
        expected = np.array(
            [
                [1.0, -1.0, -0.2724275, -0.73614431, 0.73854895, 0.0, 0.0, 0.0],
                [-1.0, 1.0, 0.2724275, 0.73614431, -0.73854895, 0.0, 0.0, 0.0],
                [-0.2724275, 0.2724275, 1.0, -0.15192004, 0.19528337, 0.0, 0.0, 0.0],
                [-0.73614431, 0.73614431, -0.15192004, 1.0, -0.87518995, 0.0, 0.0, 0.0],
                [0.73854895, -0.73854895, 0.19528337, -0.87518995, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.15430335],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.15430335, 1.0],
            ]
        )

        assert hasattr(oz, "ranks_")
        assert oz.ranks_.shape == (X.shape[1], X.shape[1])
        npt.assert_array_almost_equal(oz.ranks_, expected)

        # Image similarity comparision
        oz.finalize()
        self.assert_images_similar(oz, tol=0.1)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_rank2d_integrated_pandas(self):
        """
        Test Rank2D on occupancy dataset with pandas DataFrame and Series
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()
        features = data.meta["features"]

        assert isinstance(X, pd.DataFrame)
        assert isinstance(y, pd.Series)

        # Test the visualizer
        oz = Rank2D(features=features, show_feature_names=True)
        assert oz.fit(X, y) is oz
        assert oz.transform(X) is X
        oz.finalize()

        # Image similarity testing
        self.assert_images_similar(oz, tol=0.1)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_rank2d_integrated_numpy(self):
        """
        Test Rank2D on occupancy dataset with numpy ndarray
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()
        features = data.meta["features"]

        assert isinstance(X, np.ndarray)
        assert isinstance(y, np.ndarray)

        # Test the visualizer
        oz = Rank2D(features=features, show_feature_names=True)
        assert oz.fit(X, y) is oz
        assert oz.transform(X) is X
        oz.finalize()

        # Image similarity testing
        self.assert_images_similar(oz, tol=0.1)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_rank2d_quick_method(self):
        """
        Test Rank2D quick method
        """
        X, y = load_occupancy()
        oz = rank2d(X, y, algorithm="spearman", colormap="RdYlGn_r", show=False)

        assert isinstance(oz, Rank2D)
        self.assert_images_similar(oz, tol=0.1)



# tests.test_classifier.test_base
# Tests for the base classification visualizers
#
# Author:   Benjamin Bengfort
# Created:  Wed Jul 31 11:21:28 2019 -0400
#
# Copyright (C) 2019 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_base.py [da729da] benjamin@bengfort.com $

"""
Tests for the base classification visualizers
"""

##########################################################################
## Imports
##########################################################################

import pytest
import numpy as np
import numpy.testing as npt

from yellowbrick.classifier.base import *
from sklearn.naive_bayes import GaussianNB
from sklearn.preprocessing import LabelEncoder
from sklearn.linear_model import LinearRegression
from .conftest import assert_fitted, assert_not_fitted
from yellowbrick.exceptions import YellowbrickTypeError


##########################################################################
## Test Classification Score Visualizer
##########################################################################


@pytest.mark.usefixtures("binary", "multiclass")
class TestClassificationScoreVisualizer(object):
    """
    Test the ClassificationScoreVisualizer base functionality
    """

    def test_fit_score(self):
        """
        Ensure correct fit and score behavior
        """
        oz = ClassificationScoreVisualizer(GaussianNB())
        assert_not_fitted(oz, X_test=self.binary.X.test)
        assert oz.fit(self.binary.X.train, self.binary.y.train) is oz
        assert 0.0 <= oz.score(self.binary.X.test, self.binary.y.test) <= 1.0
        assert_fitted(oz, X_test=self.binary.X.test)

    def test_class_counts(self):
        """
        Test class and class counts identification
        """
        oz = ClassificationScoreVisualizer(GaussianNB())
        oz.fit(self.multiclass.X.train, self.multiclass.y.train)

        unique, counts = np.unique(self.multiclass.y.train, return_counts=True)
        npt.assert_array_equal(oz.classes_, unique)
        npt.assert_array_equal(oz.class_counts_, counts)

    def test_force_estimator(self):
        """
        Test that an estimator can be forced through
        """
        with pytest.raises(YellowbrickTypeError):
            ClassificationScoreVisualizer(LinearRegression())

        try:
            ClassificationScoreVisualizer(LinearRegression(), force_model=True)
        except YellowbrickTypeError as e:
            pytest.fail("type error was raised incorrectly: {}".format(e))

    def test_score_with_fitted_estimator(self):
        """
        Assert fitted estimator can be scored without fit but warns
        """
        model = GaussianNB().fit(self.binary.X.train, self.binary.y.train)

        # NOTE that the wrapper will pass a call down to `classes_`
        oz = ClassificationScoreVisualizer(model)
        assert_not_fitted(oz, ["class_counts_", "score_"])

        msg = "could not determine class_counts_"
        with pytest.warns(YellowbrickWarning, match=msg):
            oz.score(self.binary.X.test, self.binary.y.test)
            assert_fitted(oz, ["classes_", "class_counts_", "score_"])

    def test_score_without_fitted_estimator(self):
        """
        Assert score without fitted estimator raises NotFitted
        """
        oz = ClassificationScoreVisualizer(GaussianNB())
        assert_not_fitted(oz)

        with pytest.raises(NotFitted):
            oz.score(self.binary.X.test, self.binary.y.test)
            assert_not_fitted(oz)

    def test_colors_property(self):
        """
        Test that a unique color per class is created after fit
        """
        oz = ClassificationScoreVisualizer(GaussianNB())

        with pytest.raises(NotFitted, match="cannot determine colors before fit"):
            oz.class_colors_

        oz.fit(self.multiclass.X.train, self.multiclass.y.train)
        assert len(oz.class_colors_) == len(oz.classes_)

    def test_decode_labels_warning(self):
        """
        Assert warning is issued and encoder is used with multiple decoding params
        """
        with pytest.warns(
            YellowbrickWarning, match="both classes and encoder specified"
        ):
            oz = ClassificationScoreVisualizer(
                GaussianNB(),
                classes=["a", "b", "c"],
                encoder={0: "foo", 1: "bar", 2: "zap"},
            )
            encoded = oz._decode_labels([0, 1, 2])
            npt.assert_array_equal(encoded, ["foo", "bar", "zap"])

    def test_decode_labels_from_numeric(self):
        """
        Test that a numeric y can be decoded using classes and encoder
        """
        classes = np.array(["a", "b", "c", "d", "e"])
        y = np.random.randint(0, 5, 100)
        decoded = classes[y]

        oz = ClassificationScoreVisualizer(GaussianNB, classes=classes)
        npt.assert_array_equal(oz._decode_labels(y), decoded)

        encoder = dict(zip(range(len(classes)), classes))
        oz = ClassificationScoreVisualizer(GaussianNB, encoder=encoder)
        npt.assert_array_equal(oz._decode_labels(y), decoded)

        encoder = LabelEncoder().fit(decoded)
        oz = ClassificationScoreVisualizer(GaussianNB, encoder=encoder)
        npt.assert_array_equal(oz._decode_labels(y), decoded)

    def test_decode_labels_from_strings(self):
        """
        Test that string y can be decoded using classes and encoder
        """
        classes = np.array(["a", "b", "c", "d", "e"])
        decoded = classes[np.random.randint(0, 5, 100)]
        y = np.array([v.upper() for v in decoded])

        oz = ClassificationScoreVisualizer(GaussianNB, classes=classes)
        npt.assert_array_equal(oz._decode_labels(y), decoded)

        encoder = {c.upper(): c for c in classes}
        oz = ClassificationScoreVisualizer(GaussianNB, encoder=encoder)
        npt.assert_array_equal(oz._decode_labels(y), decoded)

        class L2UTransformer(object):
            def transform(self, y):
                return np.array([yi.upper() for yi in y])

            def inverse_transform(self, y):
                return np.array([yi.lower() for yi in y])

        oz = ClassificationScoreVisualizer(GaussianNB, encoder=L2UTransformer())
        npt.assert_array_equal(oz._decode_labels(y), decoded)

    def test_decode_labels_unknown_class(self):
        """
        Ensure a human-understandable error is raised when decode fails
        """
        classes = np.array(["a", "b", "c", "d", "e"])
        y = classes[np.random.randint(0, 5, 100)]

        # Remove class "c" from the known array labels
        classes = np.array(["a", "b", "d", "e"])

        oz = ClassificationScoreVisualizer(GaussianNB, classes=classes)
        with pytest.raises(ModelError, match="could not decode"):
            npt.assert_array_equal(oz._decode_labels(y), decoded)

        encoder = dict(zip(classes, range(len(classes))))
        oz = ClassificationScoreVisualizer(GaussianNB, encoder=encoder)
        with pytest.raises(
            ModelError,
            match=r"cannot decode class (np\.str_\()?\s*'c'\s*(\))? to label",
        ):
            npt.assert_array_equal(oz._decode_labels(y), decoded)

        encoder = LabelEncoder().fit(classes[np.random.randint(0, 4, 100)])
        oz = ClassificationScoreVisualizer(GaussianNB, encoder=encoder)
        with pytest.raises(ModelError, match="could not decode"):
            npt.assert_array_equal(oz._decode_labels(y), decoded)

    def test_labels(self):
        """
        Check visualizer can return human labels correctly
        """
        classes = np.array(["a", "b", "c", "d", "e"])
        y = classes[np.random.randint(0, 5, 100)]

        oz = ClassificationScoreVisualizer(GaussianNB, classes=classes)
        npt.assert_array_equal(oz._labels(), classes)

        encoder = dict(zip(range(len(classes)), classes))
        oz = ClassificationScoreVisualizer(GaussianNB, encoder=encoder)
        npt.assert_array_equal(oz._labels(), classes)

        encoder = LabelEncoder().fit(y)
        oz = ClassificationScoreVisualizer(GaussianNB, encoder=encoder)
        npt.assert_array_equal(oz._labels(), classes)

    def test_labels_warning(self):
        """
        Assert warning and encoder is used with multiple decoding params for labels
        """
        with pytest.warns(
            YellowbrickWarning, match="both classes and encoder specified"
        ):
            oz = ClassificationScoreVisualizer(
                GaussianNB(),
                classes=["a", "b", "c"],
                encoder={0: "foo", 1: "bar", 2: "zap"},
            )
            labels = oz._labels()
            npt.assert_array_equal(labels, ["foo", "bar", "zap"])

    def test_labels_encoder_no_classes(self):
        """
        Assert warning and None returned if encoder doesn't have classes
        """

        class L2UTransformer(object):
            def transform(self, y):
                return np.array([yi.upper() for yi in y])

        oz = ClassificationScoreVisualizer(GaussianNB(), encoder=L2UTransformer())
        with pytest.warns(YellowbrickWarning, match="could not determine class labels"):
            assert oz._labels() is None

    def test_dict_labels_sorted(self):
        """
        Ensure dictionary encoder labels are returned sorted
        """
        le = {3: "a", 2: "c", 1: "b"}
        oz = ClassificationScoreVisualizer(GaussianNB(), encoder=le)
        npt.assert_array_equal(oz._labels(), ["b", "c", "a"])



# tests.test_classifier.test_threshold
# Ensure that the discrimination threshold visualizations work.
#
# Author:  Nathan Danielsen
# Author:  Benjamin Bengfort
# Created: Wed April 26 20:17:29 2017 -0700
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_threshold.py [] nathan.danielsen@gmail.com $

"""
Ensure that the DiscriminationThreshold visualizations work.
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest
import yellowbrick as yb
import matplotlib.pyplot as plt

from yellowbrick.classifier.threshold import *
from yellowbrick.datasets import load_occupancy, load_spam
from yellowbrick.utils import is_probabilistic, is_classifier

from unittest.mock import patch
from tests.base import VisualTestCase
from numpy.testing import assert_array_equal

from sklearn.base import ClassifierMixin
from sklearn.svm import LinearSVC, NuSVC
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier
from sklearn.naive_bayes import BernoulliNB, GaussianNB
from sklearn.linear_model import Ridge, LogisticRegression
from sklearn.model_selection import StratifiedShuffleSplit

try:
    import pandas as pd
except ImportError:
    pd = None

##########################################################################
## DiscriminationThreshold Test Cases
##########################################################################


class TestDiscriminationThreshold(VisualTestCase):
    """
    DiscriminationThreshold visualizer tests
    """

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_binary_discrimination_threshold(self):
        """
        Correctly generates viz for binary classification with BernoulliNB
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=854,
        )

        _, ax = plt.subplots()

        model = BernoulliNB()
        visualizer = DiscriminationThreshold(model, ax=ax, random_state=23)

        visualizer.fit(X, y)
        visualizer.finalize()

        self.assert_images_similar(visualizer, tol=0.012)

    def test_multiclass_discrimination_threshold(self):
        """
        Assert exception is raised in multiclass case.
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=3,
            n_clusters_per_class=4,
            random_state=854,
        )

        visualizer = DiscriminationThreshold(GaussianNB(), random_state=23)
        msg = "multiclass format is not supported"

        with pytest.raises(ValueError, match=msg):
            visualizer.fit(X, y)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_pandas_integration(self):
        """
        Test with Pandas DataFrame and Series input
        """
        _, ax = plt.subplots()

        # Load the occupancy dataset from fixtures
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()

        classes = ["unoccupied", "occupied"]

        # Create the visualizer
        viz = DiscriminationThreshold(
            LogisticRegression(), ax=ax, classes=classes, random_state=193
        )
        viz.fit(X, y)
        viz.finalize()

        self.assert_images_similar(viz, tol=0.1)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_numpy_integration(self):
        """
        Test with NumPy arrays
        """
        _, ax = plt.subplots()

        # Load the occupancy dataset from fixtures
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()

        classes = ["unoccupied", "occupied"]

        # Create the visualizer
        viz = DiscriminationThreshold(
            LogisticRegression(), ax=ax, classes=classes, random_state=193
        )
        viz.fit(X, y)
        viz.finalize()

        self.assert_images_similar(viz, tol=0.1)

    def test_quick_method(self):
        """
        Test for thresholdviz quick method with random dataset
        """

        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=2721,
        )

        _, ax = plt.subplots()

        discrimination_threshold(BernoulliNB(), X, y, ax=ax, random_state=5, show=False)
        self.assert_images_similar(ax=ax, tol=10)

    @patch.object(DiscriminationThreshold, "draw", autospec=True)
    def test_fit(self, mock_draw):
        """
        Test the fit method generates scores, calls draw, and returns self
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=1221,
        )

        visualizer = DiscriminationThreshold(BernoulliNB())
        assert not hasattr(visualizer, "thresholds_")
        assert not hasattr(visualizer, "cv_scores_")

        out = visualizer.fit(X, y)

        assert out is visualizer
        mock_draw.assert_called_once()
        assert hasattr(visualizer, "thresholds_")
        assert hasattr(visualizer, "cv_scores_")

        for metric in METRICS:
            assert metric in visualizer.cv_scores_
            assert "{}_lower".format(metric) in visualizer.cv_scores_
            assert "{}_upper".format(metric) in visualizer.cv_scores_

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_binary_discrimination_threshold_alt_args(self):
        """
        Correctly generates visualization with alternate arguments
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=10,
            n_redundant=3,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=1231,
            flip_y=0.1,
            weights=[0.35, 0.65],
        )

        exclude = ["queue_rate", "fscore"]
        cv = StratifiedShuffleSplit(n_splits=1, test_size=0.2)
        visualizer = DiscriminationThreshold(
            NuSVC(), exclude=exclude, cv=cv, random_state=98239
        )

        visualizer.fit(X, y)
        visualizer.finalize()

        for metric in exclude:
            assert metric not in visualizer.cv_scores_
            assert "{}_lower".format(metric) not in visualizer.cv_scores_
            assert "{}_upper".format(metric) not in visualizer.cv_scores_

        self.assert_images_similar(visualizer)

    def test_threshold_default_initialization(self):
        """
        Test initialization default parameters
        """
        model = BernoulliNB()
        viz = DiscriminationThreshold(model)
        assert viz.estimator is model
        assert viz.color is None
        assert viz.title is None
        assert viz.n_trials == 50
        assert viz.cv == 0.1
        assert_array_equal(viz.quantiles, np.array((0.1, 0.5, 0.9)))

    def test_requires_classifier(self):
        """
        Assert requires a classifier
        """
        message = "requires a probabilistic binary classifier"
        assert not is_classifier(Ridge)

        with pytest.raises(yb.exceptions.YellowbrickError, match=message):
            DiscriminationThreshold(Ridge())

    def test_requires_probabilistic_classifier(self):
        """
        Assert requires probabilistic classifier
        """
        message = "requires a probabilistic binary classifier"

        class RoboClassifier(ClassifierMixin):
            """
            Dummy Non-Probabilistic Classifier
            """

            def fit(self, X, y):
                self.classes_ = [0, 1]
                return self

        assert is_classifier(RoboClassifier)
        assert not is_probabilistic(RoboClassifier)

        with pytest.raises(yb.exceptions.YellowbrickError, match=message):
            DiscriminationThreshold(RoboClassifier())

    def test_accepts_predict_proba(self):
        """
        Will accept classifiers with predict proba function
        """
        model = RandomForestClassifier
        assert is_classifier(model)
        assert is_probabilistic(model)
        assert not hasattr(model, "decision_function")
        assert hasattr(model, "predict_proba")

        try:
            DiscriminationThreshold(model())
        except YellowbrickTypeError:
            pytest.fail("did not accept decision function model")

    def test_accepts_decision_function(self):
        """
        Will accept classifiers with decision function
        """
        model = LinearSVC
        assert is_classifier(model)
        assert is_probabilistic(model)
        assert hasattr(model, "decision_function")
        assert not hasattr(model, "predict_proba")

        try:
            DiscriminationThreshold(model())
        except YellowbrickTypeError:
            pytest.fail("did not accept decision function model")

    def test_bad_quantiles(self):
        """
        Assert exception is raised when bad quantiles are passed in.
        """
        msg = (
            "quantiles must be a sequence of three "
            "monotonically increasing values less than 1"
        )

        with pytest.raises(YellowbrickValueError, match=msg):
            DiscriminationThreshold(NuSVC(), quantiles=[0.25, 0.1, 0.75])

    def test_bad_cv(self):
        """
        Assert an exception is raised when a bad cv value is passed in
        """
        with pytest.raises(YellowbrickValueError, match="not a valid cv splitter"):
            DiscriminationThreshold(NuSVC(), cv="foo")

    def test_splitter_random_state(self):
        """
        Test splitter random state is modified
        """
        viz = DiscriminationThreshold(NuSVC(), random_state=None)
        assert viz._check_cv(None, random_state=None).random_state is None
        assert viz._check_cv(None, random_state=42).random_state == 42

        splits = StratifiedShuffleSplit(n_splits=1, random_state=None)
        assert viz._check_cv(splits, random_state=None).random_state is None
        assert viz._check_cv(splits, random_state=23).random_state == 23

        splits = StratifiedShuffleSplit(n_splits=1, random_state=181)
        assert viz._check_cv(splits, random_state=None).random_state == 181
        assert viz._check_cv(splits, random_state=72).random_state == 72

    def test_bad_exclude(self):
        """
        Assert an exception is raised on bad exclude param
        """
        with pytest.raises(YellowbrickValueError, match="not a valid metric"):
            DiscriminationThreshold(NuSVC(), exclude="foo")

        with pytest.raises(YellowbrickValueError, match="not a valid metric"):
            DiscriminationThreshold(NuSVC(), exclude=["queue_rate", "foo"])

    def test_bad_argmax(self):
        """
        Assert an exception is raised on bad argmax param
        """
        with pytest.raises(YellowbrickValueError, match="not a valid metric"):
            DiscriminationThreshold(NuSVC(), argmax="foo")

    @pytest.mark.parametrize(
        "viz_kwargs",
        [
            {"argmax": None},
            {"exclude": "fscore"},
            {"argmax": "queue_rate", "exclude": "queue_rate"},
        ],
    )
    def test_none_argmax(self, viz_kwargs):
        """
        Assert argmax param works fine
        """
        viz = DiscriminationThreshold(NuSVC(), **viz_kwargs)
        assert viz._check_argmax(viz.argmax, viz.exclude) is None

    def test_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_spam()

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                (
                    "dt",
                    DiscriminationThreshold(
                        LogisticRegression(multi_class="auto", solver="liblinear"),
                        random_state=42,
                    ),
                ),
            ]
        )

        model.fit(X, y)
        model["dt"].finalize()
        self.assert_images_similar(model["dt"], tol=15)

    def test_within_pipeline_quickmethod(self):
        """
        Test that visualizer quickmethod can be accessed within a
        sklearn pipeline
        """
        X, y = load_spam()

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                (
                    "dt",
                    discrimination_threshold(
                        LogisticRegression(multi_class="auto", solver="liblinear"),
                        X,
                        y,
                        random_state=42,
                    ),
                ),
            ]
        )
        model["dt"].finalize()
        self.assert_images_similar(model["dt"], tol=15)

    def test_pipeline_as_model_input(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        """
        X, y = load_spam()

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                ("lr", LogisticRegression(multi_class="auto", solver="liblinear")),
            ]
        )

        oz = DiscriminationThreshold(model, random_state=42)
        oz.fit(X, y)
        oz.finalize()
        self.assert_images_similar(oz, tol=15)

    def test_pipeline_as_model_input_quickmethod(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        within a quickmethod
        """
        X, y = load_spam()

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                ("lr", LogisticRegression(multi_class="auto", solver="liblinear")),
            ]
        )

        oz = discrimination_threshold(model, X, y, random_state=42)
        self.assert_images_similar(oz, tol=15)



# tests.test_datasets.test_paths
# Tests for the dataset path utilities
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Thu Jul 26 14:28:14 2018 -0400
#
# ID: test_path.py [7082742] benjamin@bengfort.com $

"""
Tests for the dataset path utilities
"""

##########################################################################
## Imports
##########################################################################

import os
import pytest
import contextlib

from yellowbrick.datasets.path import *
from yellowbrick.exceptions import DatasetsError


##########################################################################
## Utilities
##########################################################################


@contextlib.contextmanager
def environ(**env):
    """
    Temporarily set the environment variables for a test, restoring them when
    the test is complete (e.g. when the context manager exits).

    Parameters
    ----------
    env : dict
        The environment variables that should exist in context
    """
    old_env = dict(os.environ)
    os.environ.clear()
    os.environ.update(env)

    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_env)


##########################################################################
## Test Cases
##########################################################################


def test_get_data_home_fixtures():
    """
    get_data_home should return fixtures by default
    """
    assert get_data_home() == FIXTURES
    assert os.path.exists(FIXTURES)


def test_get_data_home_env(tmpdir):
    """
    get_data_home should return the environment variable if set
    """
    path = str(tmpdir.mkdir("fixtures").join("foo"))
    assert not os.path.exists(path)

    with environ(YELLOWBRICK_DATA=path):
        assert get_data_home() == path
        assert os.path.exists(path)


def test_get_data_home_specified(tmpdir):
    """
    get_data_home should return a passed in path
    """
    path = str(tmpdir.mkdir("fixtures").join("foo"))
    assert not os.path.exists(path)

    assert get_data_home(path) == path
    assert os.path.exists(path)


def test_find_dataset_path(tmpdir):
    """
    Test find_dataset_path with a specified data_home
    """

    # Create the dataset
    data_home = tmpdir.mkdir("fixtures")
    foo = data_home.mkdir("foo")

    # Test the default lookup of foo/foo.csv.gz
    fpath = foo.join("foo.csv.gz")
    fpath.write("1,2,3")
    assert find_dataset_path("foo", data_home=data_home) == fpath

    # Test the extension based lookup of foo/foo.npz
    fpath = foo.join("foo.npz")
    fpath.write("1234")
    assert find_dataset_path("foo", data_home=data_home, ext=".npz") == fpath

    # Test the fname based lookup of foo/data.txt
    fpath = foo.join("data.txt")
    fpath.write("there is data in this file")
    assert find_dataset_path("foo", data_home=data_home, fname="data.txt") == fpath


def test_missing_find_dataset_path(tmpdir):
    """
    Test find_dataset_path when the dataset does not exist
    """
    data_home = tmpdir.mkdir("fixtures")

    # When the data directory doesn't exist
    with pytest.raises(DatasetsError):
        find_dataset_path("foo", data_home=str(data_home))

    # When the data directory exists but no file is in the directory
    foo = data_home.mkdir("foo")
    with pytest.raises(DatasetsError):
        find_dataset_path("foo", data_home=str(data_home))

    # When the specified file doesn't exist
    fpath = foo.join("foo.csv")
    fpath.write("1,2,3")
    with pytest.raises(DatasetsError):
        find_dataset_path("foo", data_home=str(data_home), ext=".npz")


def test_suppress_exception_find_dataset_path(tmpdir):
    """
    Assert that find_dataset_path can suppress exceptions
    """
    data_home = str(tmpdir.mkdir("fixtures"))
    assert find_dataset_path("foo", data_home=data_home, raises=False) is None


def test_dataset_exists(tmpdir):
    """
    Test the dataset_exists helper function
    """
    data_home = tmpdir.mkdir("fixtures")
    assert not os.path.exists(str(data_home.join("foo")))

    # Test when directory doesn't exist
    assert not dataset_exists("foo", str(data_home))

    # Test when path exists but is file
    fpath = data_home.join("foo.txt")
    fpath.write("foo")

    assert not dataset_exists("foo.txt", str(data_home))

    # Test correct case
    data_home.mkdir("foo")
    assert dataset_exists("foo", str(data_home))


def test_dataset_archive(tmpdir):
    """
    Test the dataset_archive determines if an archive is up to date
    """
    sig = "49b3fc3143d727d7819fabd4365d7e7b29794089dc9fa1e5e452aeb0b33d5eda"
    data_home = tmpdir.mkdir("fixtures")

    # When archive does not exist
    assert not dataset_archive("foo", sig, data_home=str(data_home))

    # Create archive
    fpath = data_home.join("foo.zip")
    fpath.write("this is a data archive")

    # When archive exists
    assert dataset_archive("foo", sig, data_home=data_home)

    # When archive does not match signature
    assert not dataset_archive("foo", "abcd", data_home=data_home)


def test_cleanup_dataset(tmpdir):
    """
    Test cleanup_dataset removes both data dir and archive
    """
    data_home = tmpdir.mkdir("fixtures")

    # Make dataset and archive
    foo = data_home.mkdir("foo")
    fdata = foo.join("foo.csv")
    fdata.write("testing 1,2,3")

    fzip = data_home.join("foo.zip")
    fzip.write("this is the archive file")

    # Make sure the files exist
    assert os.path.exists(fzip)
    assert os.path.exists(fdata)

    # Cleanup the dataset
    cleanup_dataset("foo", data_home=data_home)

    # Files should be gone
    assert not os.path.exists(fzip)
    assert not os.path.exists(fdata)


def test_cleanup_dataset_no_data(tmpdir):
    """
    Assert cleanup_dataset fails gracefully if data and archive don't exist.
    """
    data_home = tmpdir.mkdir("fixtures")
    cleanup_dataset("foo", data_home=str(data_home))

    # Files should be gone
    assert not os.path.exists(str(data_home.join("foo.zip")))
    assert not os.path.exists(str(data_home.join("foo")))



# tests.random
# A visualizer that draws a random scatter plot for testing.
#
# Author:  Benjamin Bengfort <bbengfort@districtdatalabs.com>
# Created: Wed Mar 21 17:51:15 2018 -0400
#
# ID: rand.py [cc69b3c] davidwaterman@gmail.com $

"""
A visualizer that draws a random scatter plot for testing.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from yellowbrick.base import Visualizer
from yellowbrick.style import resolve_colors

from sklearn.datasets import make_blobs


##########################################################################
## Random Visualizer
##########################################################################


class RandomVisualizer(Visualizer):
    """
    Creates random scatter plots as a testing utility.

    Data generation uses scikit-learn make_blobs to create scatter plots that
    have reasonable visual features and multiple colors.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    n_samples : int, default: 100
        The number of points to generate for the scatter plot

    n_blobs : int or array of shape [n_centers, 2]
        Define the number of blobs to create or specify their centers.

    random_state : int, RandomState or None:
        Used to specify the seed of the random state to ensure tests work.
    """

    def __init__(self, ax=None, n_samples=100, n_blobs=3, random_state=None, **kwargs):

        super(RandomVisualizer, self).__init__(ax=ax, **kwargs)
        if isinstance(random_state, (int, float)) or random_state is None:
            random_state = np.random.RandomState(random_state)

        self.n_samples = n_samples
        self.n_blobs = n_blobs
        self.random_state = random_state

    def generate(self):
        """
        Returns random data according to the visualizer specification.

        Returns
        -------
        X : array of shape [n_samples, 2]
            2 dimensional array of points to plot

        y : vector with length n_samples
            Center/blob each point belongs to (used for color)
        """
        return make_blobs(
            n_samples=self.n_samples, n_features=2, centers=self.n_blobs, random_state=self.random_state
        )

    def fit(self, *args, **kwargs):
        X, c = self.generate()

        x = X[:, 0]
        y = X[:, 1]

        self.draw(x, y, c)
        return self

    def draw(self, x, y, c):
        colors = resolve_colors(self.n_blobs)

        for i in np.arange(self.n_blobs):
            mask = c == i
            label = "c{}".format(i)
            self.ax.scatter(x[mask], y[mask], label=label, c=colors[i])

        return self.ax

    def finalize(self):
        self.ax.legend(frameon=True)
        self.ax.set_ylabel("$y$")
        self.ax.set_xlabel("$x$")
        self.ax.set_title("Random Scatter Plot")
        return self.ax


if __name__ == "__main__":
    r = RandomVisualizer()
    r.fit()
    r.show(outpath="test.png")



# tests.test_target.test_class_balance
# Tests for the ClassBalance visualizer
#
# Author:  Benjamin Bengfort
# Created: Thu Jul 19 10:21:49 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_class_balance.py [d742c57] benjamin@bengfort.com $

"""
Tests for the ClassBalance visualizer
"""

##########################################################################
## Imports
##########################################################################

import pytest

from yellowbrick.target.class_balance import *
from yellowbrick.datasets import load_occupancy
from yellowbrick.exceptions import YellowbrickValueError

from tests.fixtures import Dataset, Split
from tests.base import VisualTestCase

from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split as tts

try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
# Data Fixtures
##########################################################################

# TODO: convert to Pytest fixture
def make_fixture(binary=False, balanced=False, split=False):
    """
    Make a dataset for testing ClassBalance based on the specified params.
    """
    kwargs = {
        "n_samples": 100,
        "n_features": 20,
        "n_informative": 8,
        "n_redundant": 2,
        "n_clusters_per_class": 1,
        "random_state": 89092,
    }

    if binary:
        kwargs["n_classes"] = 2
        kwargs["weights"] = None if balanced else [0.3, 0.7]
    else:
        kwargs["n_classes"] = 5
        kwargs["weights"] = None if balanced else [0.1, 0.2, 0.4, 0.2, 0.01]

    X, y = make_classification(**kwargs)

    if split:
        X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=101)
        return Dataset(Split(X_train, X_test), Split(y_train, y_test))

    return Dataset(X, y)


##########################################################################
#  Tests
##########################################################################


class TestClassBalance(VisualTestCase):
    """
    Test ClassBalance visualizer
    """

    def test_signature_exception(self):
        """
        An exception is raised if X and y are put into the visualizer
        """
        oz = ClassBalance()
        dataset = make_fixture(split=False)

        message = "fit has changed to only require a 1D array, y"
        with pytest.raises(YellowbrickValueError, match=message):
            oz.fit(dataset.X, dataset.y)

    def test_invalid_target(self):
        """
        A value error should be raised on invalid train or test target
        """
        y_valid = np.random.randint(2, size=100)
        y_invalid = np.random.uniform(size=100)

        oz = ClassBalance()

        with pytest.raises(YellowbrickValueError):
            oz.fit(y_invalid)

        with pytest.raises(YellowbrickValueError):
            oz.fit(y_valid, y_invalid)

    def test_class_names_must_match(self):
        """
        Assert error raised when more classes are in data than specified
        """
        oz = ClassBalance(labels=["a", "b", "c"])
        dataset = make_fixture(binary=False, split=False)

        with pytest.raises(YellowbrickValueError):
            oz.fit(dataset.y)

    def test_binary_balance(self):
        """
        Test binary classification in balance mode
        """
        dataset = make_fixture(binary=True, split=False)

        oz = ClassBalance()
        assert oz.fit(dataset.y) is oz
        assert oz._mode == BALANCE

        # oz.finalize()
        self.assert_images_similar(oz)

    def test_binary_compare(self):
        """
        Test binary classification in compare mode
        """
        dataset = make_fixture(binary=True, split=True)

        oz = ClassBalance()
        assert oz.fit(dataset.y.train, dataset.y.test) is oz
        assert oz._mode == COMPARE

        # oz.finalize()
        self.assert_images_similar(oz)

    def test_multiclass_balance(self):
        """
        Test multiclass classification in balance mode
        """
        dataset = make_fixture(binary=False, split=False)

        oz = ClassBalance()
        assert oz.fit(dataset.y) is oz
        assert oz._mode == BALANCE

        # oz.finalize()
        self.assert_images_similar(oz)

    def test_multiclass_compare(self):
        """
        Test multiclass classification in compare mode
        """
        dataset = make_fixture(binary=False, split=True)

        oz = ClassBalance()
        assert oz.fit(dataset.y.train, dataset.y.test) is oz
        assert oz._mode == COMPARE

        # oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_pandas_occupancy_balance(self):
        """
        Test pandas data frame with string target in balance mode
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()

        # Create and fit the visualizer
        oz = ClassBalance()
        assert oz.fit(y) is oz

        # oz.finalize()
        self.assert_images_similar(oz)

    def test_numpy_occupancy_balance(self):
        """
        Test NumPy arrays with string target in balance mode
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()

        # Create and fit the visualizer
        oz = ClassBalance()
        assert oz.fit(y) is oz

        # oz.finalize()
        self.assert_images_similar(oz)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_pandas_occupancy_compare(self):
        """
        Test pandas data frame with string target in compare mode
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()

        _, _, y_train, y_test = tts(X, y, test_size=0.4, random_state=2242)

        # Create and fit the visualizer
        oz = ClassBalance()
        assert oz.fit(y_train, y_test) is oz

        # oz.finalize()
        self.assert_images_similar(oz, tol=0.5)  # w/o tol fails with RMS 0.433

    def test_numpy_occupancy_compare(self):
        """
        Test NumPy arrays with string target in compare mode
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()

        _, _, y_train, y_test = tts(X, y, test_size=0.4, random_state=2242)

        # Create and fit the visualizer
        oz = ClassBalance()
        assert oz.fit(y_train, y_test) is oz

        # oz.finalize()
        self.assert_images_similar(oz, tol=0.5)  # w/o tol fails with RMS 0.433

    def test_quick_method(self):
        """
        Test the quick method producing a valid visualization
        """
        dataset = make_fixture(binary=False, split=False)

        viz = class_balance(dataset.y, show=False)

        assert isinstance(viz, ClassBalance)
        self.assert_images_similar(viz, tol=0.5)

    def test_quick_method_with_splits(self):
        """
        Test the quick method works with train and test splits
        """
        dataset = make_fixture(binary=False, split=True)

        viz = class_balance(dataset.y.train, dataset.y.test, show=False)

        assert isinstance(viz, ClassBalance)
        self.assert_images_similar(viz)
